"""Useful functions for the Singularity containers

TODO:
- [x] figure out how to mount in other file-systems
  -B dir1,dir2

Put to release notes:
`conda install -c bioconda singularity`

OR

`conda install -c conda-forge singularity`

"""
from __future__ import absolute_import
from __future__ import print_function

import six
import os
from kipoi_utils.utils import unique_list, makedir_exist_ok, is_subdir
from kipoi_conda import _call_command
import subprocess
import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


# Python wrapper for the Singularity CLI
# def assert_installed():
#     """Make sure singularity is installed
#     """
#     pass


def singularity_pull(remote_path, local_path):
    """Run `singularity pull`

    Args:
      remote_path: singularity remote path. Example: shub://kipoi/models:latest
      local_path: local file path to the ".sif" file
    """
    makedir_exist_ok(os.path.dirname(local_path))
    if os.path.exists(local_path):
        logger.info("Container file {} already exists. Skipping `singularity pull`".
                    format(local_path))
    else:
        if os.environ.get('SINGULARITY_CACHEDIR'):
            downloaded_path = os.path.join(os.environ.get('SINGULARITY_CACHEDIR'),
                                           os.path.basename(local_path))
            pull_dir = os.path.dirname(downloaded_path)
            logger.info("SINGULARITY_CACHEDIR is set to {}".
                        format(os.environ.get('SINGULARITY_CACHEDIR')))
            if os.path.exists(downloaded_path):
                logger.info("Container file {} already exists. Skipping `singularity pull` and softlinking it".
                            format(downloaded_path))
                if os.path.islink(local_path):
                    logger.info("Softlink {} already exists. Removing it".format(local_path))
                    os.remove(local_path)

                logger.info("Soflinking the downloaded file: ln -s {} {}".
                            format(downloaded_path,
                                   local_path))
                os.symlink(downloaded_path, local_path)
                return None
        else:
            pull_dir = os.path.dirname(local_path)

        logger.info("Container file {} doesn't exist. Pulling the container from {}. Saving it to: {}".
                    format(local_path, remote_path, pull_dir))
        cmd = ['singularity', 'pull', '--name', os.path.basename(local_path), remote_path]
        logger.info(" ".join(cmd))
        returncode = subprocess.call(cmd,
                                     cwd=pull_dir)
        if returncode != 0:
            raise ValueError("Command: {} failed".format(" ".join(cmd)))

        # softlink it
        if os.environ.get('SINGULARITY_CACHEDIR'):
            if os.path.islink(local_path):
                logger.info("Softlink {} already exists. Removing it".format(local_path))
                os.remove(local_path)
            logger.info("Soflinking the downloaded file: ln -s {} {}".
                        format(downloaded_path,
                               local_path))
            os.symlink(downloaded_path, local_path)

        if not os.path.exists(local_path):
            raise ValueError("Container doesn't exist at the download path: {}".format(local_path))


def singularity_exec(container, command, bind_directories=[], dry_run=False):
    """Run `singularity exec`

    Args:
      container: path to the singularity image (*.sif)
      command: command to run (as a list)
      bind_directories: Additional directories to bind
    """
    if bind_directories:
        options = ['-B', ",".join(bind_directories)]
    else:
        options = []

    cmd = ['singularity', 'exec'] + options + [container] + command
    logger.info(" ".join(cmd))
    if dry_run:
        return print(" ".join(cmd))
    else:
        returncode = subprocess.call(cmd,
                                     stdin=subprocess.PIPE)
    if returncode != 0:
        raise ValueError("Command: {} failed".format(" ".join(cmd)))


# --------------------------------------------
# Figure out relative paths:
# - container path (e.g. shub://kipoi/models:latest)
# - local path (e.g. ~/.kipoi/envs/singularity/kipoi/models_latest.sif)

def container_remote_url(source='kipoi'):
    if source == 'kipoi':
        return 'shub://kipoi/models:latest'
    else:
        raise NotImplementedError("Containers for sources other than Kipoi are not yet implemented")


def container_local_path(remote_path):
    from kipoi.config import _kipoi_dir
    tmp = os.path.join(remote_path.split("://")[1])
    if ":" in tmp:
        relative_path, tag = tmp.split(":")
    else:
        relative_path = tmp
        tag = 'latest'
    return os.path.join(_kipoi_dir, "envs/singularity/", relative_path + "_" + tag + ".sif")

# ---------------------------------


def involved_directories(dataloader_kwargs, output_files=[], exclude_dirs=[]):
    """Infer the involved directories given dataloader kwargs
    """
    dirs = []
    # dataloader kwargs
    for k, v in six.iteritems(dataloader_kwargs):
        if os.path.exists(v):
            dirs.append(os.path.dirname(os.path.abspath(v)))

    # output files
    for v in output_files:
        dirs.append(os.path.dirname(os.path.abspath(v)))

    # optionally exclude directories
    def in_any_dir(fname, dirs):
        return any([is_subdir(fname, os.path.expanduser(d))
                    for d in dirs])
    dirs = [x for x in dirs
            if not in_any_dir(x, exclude_dirs)]

    return unique_list(dirs)


def create_conda_run():
    """Create conda_run bash script to ~/.kipoi/bin/conda_run

    NOTE: this should be changed to `conda run` once conda=4.6.0 is released
    https://github.com/conda/conda/issues/2379
    """
    from kipoi.config import _kipoi_dir
    crun = """#!/bin/bash
# Run a bash command in a new conda environment
set -e # stop on error

if [[ $# -lt 2 ]] ; then
    echo "Usage: "
    echo "       conda_run <conda envrionment> <command> "
    exit 0
fi

env=$1
cmd=${@:2}
echo "Running command in env: $env"
echo "Command: $cmd"

source activate $env
$cmd
source deactivate $env
"""
    bin_dir = os.path.join(_kipoi_dir, 'bin')
    makedir_exist_ok(bin_dir)
    crun_path = os.path.join(bin_dir, 'conda_run')
    with open(crun_path, 'w') as f:
        f.write(crun)

    # make it executable
    subprocess.call(["chmod", "u+x", crun_path])
    return crun_path


def singularity_command(kipoi_cmd, model, dataloader_kwargs, output_files=[], source='kipoi', dry_run=False):

    remote_path = container_remote_url(source)
    local_path = container_local_path(remote_path)
    singularity_pull(remote_path, local_path)

    assert kipoi_cmd[0] == 'kipoi'

    # remove all spaces within each command
    kipoi_cmd = [x.replace(" ", "").replace("\n", "").replace("\t", "") for x in kipoi_cmd]

    # figure out the right environment name
    stdout, stderr = _call_command('singularity', ['exec', local_path, 'kipoi', 'env', 'get', model], stdin=subprocess.PIPE)
    env_name = stdout.decode().strip()

    # create/get the `conda_run` command
    conda_run = create_conda_run()

    singularity_exec(local_path,
                     [conda_run, env_name] + kipoi_cmd,
                     # kipoi_cmd_conda,
                     bind_directories=involved_directories(dataloader_kwargs, output_files, exclude_dirs=['/tmp', '~']), dry_run=dry_run)



"""Defines the classes for Yaml parsing using related: https://github.com/genomoncology/related
"""
from __future__ import absolute_import
from __future__ import print_function

import collections
import logging
import os
from collections import OrderedDict

import enum
import numpy as np
import related
import six

import kipoi
from kipoi_utils.external.torchvision.dataset_utils import download_url, check_integrity
from kipoi.plugin import get_model_yaml_parser, get_dataloader_yaml_parser, is_installed
import kipoi_conda as kconda
from kipoi_utils.external.related.fields import StrSequenceField, NestedMappingField, TupleIntField, AnyField, \
    UNSPECIFIED
from kipoi_utils.external.related.mixins import RelatedConfigMixin, RelatedLoadSaveMixin
from kipoi.metadata import GenomicRanges
from kipoi_utils.utils import (unique_list, yaml_ordered_dump, read_txt, read_yaml,
                               load_obj, inherits_from, override_default_kwargs, recursive_dict_parse)

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


# --------------------------------------------
# Common specs (model and dataloader)


@related.immutable(strict=False)
class Author(RelatedConfigMixin):
    name = related.StringField()
    github = related.StringField(required=False)
    email = related.StringField(required=False)


@related.mutable(strict=False)
class Info(RelatedConfigMixin):
    """Class holding information about the component.
    Parses the info section in component.yaml:

    info:
      authors:
        - name: Ziga Avsec
      doc: RBP binding prediction
      name: rbp_eclip
      version: 0.1
    """
    authors = related.SequenceField(Author, repr=True, required=False)
    doc = related.StringField("", required=False)  # free-text description of the model
    name = related.StringField(required=False)  # TODO - deprecate
    version = related.StringField(default="0.1", required=False)
    license = related.StringField(default="MIT", required=False)  # license of the model/dataloader - defaults to MIT
    tags = StrSequenceField(str, default=[], required=False)

    def __attrs_post_init__(self):
        if self.authors and self.doc == "":
            logger.warning("doc empty for the `info:` field")


@related.mutable(strict=False)
class ModelInfo(Info):
    """Additional information for the model - not applicable to the dataloader
    """
    contributors = related.SequenceField(Author, default=[], repr=True, required=False)
    cite_as = related.StringField(required=False)  # a link or a description how to cite the paper (say a doi link)
    trained_on = related.StringField(required=False)  # a link or a description of the training dataset
    training_procedure = related.StringField(
        required=False)  # brief description about the training procedure for the trained_on dataset.


@enum.unique
class ArraySpecialType(enum.Enum):
    DNASeq = "DNASeq"
    DNAStringSeq = "DNAStringSeq"
    BIGWIG = "bigwig"
    VPLOT = "v-plot"
    Array = "Array"


@related.mutable(strict=False)
class ArraySchema(RelatedConfigMixin):
    """

    Args:
      shape: Tuple of shape (same as in Keras for the input)
      doc: Description of the array
      special_type: str, special type name. Could also be an array of special entries?
      metadata_entries: str or list of metadata
    """
    verbose = True
    shape = TupleIntField()
    doc = related.StringField("", required=False)
    # MAYBE - allow a list of strings?
    #         - could be useful when a single array can have multiple 'attributes'
    name = related.StringField(required=False)
    special_type = related.ChildField(ArraySpecialType, required=False)
    associated_metadata = StrSequenceField(str, default=[], required=False)
    column_labels = StrSequenceField(str, default=[],
                                     required=False)  # either a list or a path to a file --> need to check whether it's a list

    # TODO shall we have
    # - associated_metadata in ArraySchema
    # OR
    # - associated_array in MetadataField?

    # assert that there are no Nones in the shape, assume that channels is the only 4 or it is the last
    # update the model schema shape on calling batch_iter method
    # overwrite the batch_iter method of the returned dataloader --> decorator needed

    def print_msg(self, msg):
        if self.verbose:
            print("ArraySchema mismatch")
            print(msg)

    def _validate_list_column_labels(self):
        dim_ok = len(self.shape) >= 1
        if dim_ok and (self.shape[0] is not None):
            dim_ok &= len(self.column_labels) == self.shape[0]
        if not dim_ok:
            self.print_msg("Column annotation does not match array dimension with shape %s and %d labels (%s ...)"
                           % (str(self.shape), len(self.column_labels), str(self.column_labels)[:30]))

    def __attrs_post_init__(self):
        from io import open

        if len(self.column_labels) > 1:
            # check that length is ok with columns
            self._validate_list_column_labels()
        elif len(self.column_labels) == 1:
            label = self.column_labels.list[0]
            import os
            # check if path exists raise exception only test time, but only a warning in prediction time
            if os.path.exists(label):
                with open(label, "r", encoding="utf-8") as ifh:
                    object.__setattr__(self, "column_labels", [l.rstrip() for l in ifh])
            self._validate_list_column_labels()
        else:
            object.__setattr__(self, "column_labels", None)

    def compatible_with_batch(self, batch, verbose=True):
        """Checks compatibility with a particular batch of data

        Args:
          batch: numpy array
          ignore_batch_axis: if True, the batch axis is not considered
          verbose: print the fail reason
        """

        def print_msg(msg):
            if verbose:
                print("ArraySchema mismatch")
                print(msg)

        # type = np.ndarray
        if not isinstance(batch, np.ndarray):
            print_msg("Expecting a np.ndarray. Got type(batch) = {0}".format(type(batch)))
            return False

        if not batch.ndim >= 1:
            print_msg("The array is a scalar (expecting at least the batch dimension)")
            return False

        return self.compatible_with_schema(ArraySchema(shape=batch.shape[1:],
                                                       doc=""))

    def compatible_with_schema(self, schema, name_self="", name_schema="", verbose=True):
        """Checks the compatibility with another schema

        Args:
          schema: Other ArraySchema
          name_self: How to call self in the error messages
          name_schema: analogously to name_self for the schema ArraySchema
          verbose: bool, describe what went wrong through print()
        """

        def print_msg(msg):
            if verbose:
                # print("ArraySchema mismatch")
                print(msg)

        if not isinstance(schema, ArraySchema):
            print_msg("Expecting ArraySchema. Got type({0} schema) = {1}".format(name_schema,
                                                                                 type(schema)))
            return False

        def print_msg_template():
            print("ArraySchema mismatch")
            print("Array shapes don't match for the fields:")
            print("--")
            print(name_self)
            print("--")
            print(self.get_config_as_yaml())
            print("--")
            print(name_schema)
            print("--")
            print(schema.get_config_as_yaml())
            print("--")
            print("Provided shape (without the batch axis): {0}, expected shape: {1} ".format(bshape, self.shape))

        bshape = schema.shape
        if not len(bshape) == len(self.shape):
            print_msg_template()
            return False
        for i in range(len(bshape)):
            if bshape[i] is not None and self.shape[i] is not None:
                # shapes don't match
                if not bshape[i] == self.shape[i]:
                    print_msg_template()
                    return False
        return True


# --------------------------------------------
# Model specific specs

@related.mutable(strict=False)
class ModelSchema(RelatedConfigMixin):
    """Describes the model schema
    """
    # can be a dictionary, list or a single array
    inputs = NestedMappingField(ArraySchema, keyword="shape", key="name")
    targets = NestedMappingField(ArraySchema, keyword="shape", key="name")

    def compatible_with_schema(self, dataloader_schema, verbose=True):
        """Check the compatibility: model.schema <-> dataloader.output_schema

        Checks preformed:
        - nested structure is the same (i.e. dictionary names, list length etc)
        - array shapes are compatible
        - returned obj classess are compatible

        # Arguments
            dataloader_schema: a dataloader_schema of data returned by one iteraton of dataloader's dataloader_schema_iter
                nested dictionary
            verbose: verbose error logging if things don't match

        # Returns
           bool: True only if everyhing is ok
        """

        def print_msg(msg):
            if verbose:
                print(msg)

        # Inputs check
        def compatible_nestedmapping(dschema, descr, cls, verbose=True):
            """Recursive function of checks

            shapes match, dschema-dim matches
            """
            if isinstance(descr, cls):
                # Recursion stop
                return descr.compatible_with_schema(dschema,
                                                    name_self="Model",
                                                    name_schema="Dataloader",
                                                    verbose=verbose)
            elif isinstance(dschema, collections.Mapping) and isinstance(descr, collections.Mapping):
                if not set(descr.keys()).issubset(set(dschema.keys())):
                    print_msg("Dataloader doesn't provide all the fields required by the model:")
                    print_msg("dataloader fields: {0}".format(dschema.keys()))
                    print_msg("model fields: {0}".format(descr.keys()))
                    return False
                return all([compatible_nestedmapping(dschema[key], descr[key], cls, verbose) for key in descr])
            elif isinstance(dschema, collections.Sequence) and isinstance(descr, collections.Sequence):
                if not len(descr) <= len(dschema):
                    print_msg("Dataloader doesn't provide all the fields required by the model:")
                    print_msg("len(dataloader): {0}".format(len(dschema)))
                    print_msg("len(model): {0}".format(len(descr)))
                    return False
                return all([compatible_nestedmapping(dschema[i], descr[i], cls, verbose) for i in range(len(descr))])
            elif isinstance(dschema, collections.Mapping) and isinstance(descr, collections.Sequence):
                if not len(descr) <= len(dschema):
                    print_msg("Dataloader doesn't provide all the fields required by the model:")
                    print_msg("len(dataloader): {0}".format(len(dschema)))
                    print_msg("len(model): {0}".format(len(descr)))
                    return False
                compatible = []
                for i in range(len(descr)):
                    if descr[i].name in dschema:
                        compatible.append(compatible_nestedmapping(dschema[descr[i].name], descr[i], cls, verbose))
                    else:
                        print_msg("Model array name: {0} not found in dataloader keys: {1}".
                                  format(descr[i].name, list(dschema.keys())))
                        return False
                return all(compatible)

            print_msg("Invalid types:")
            print_msg("type(Dataloader schema): {0}".format(type(dschema)))
            print_msg("type(Model schema): {0}".format(type(descr)))
            return False

        if not compatible_nestedmapping(dataloader_schema.inputs, self.inputs, ArraySchema, verbose):
            return False

        # checking targets
        if dataloader_schema.targets is None:
            return True

        if (isinstance(dataloader_schema.targets, ArraySchema) or
            len(dataloader_schema.targets) > 0) and not compatible_nestedmapping(dataloader_schema.targets,
                                                                                 self.targets,
                                                                                 ArraySchema,
                                                                                 verbose):
            return False

        return True


# --------------------------------------------
# DataLoader specific specs

@enum.unique
class MetadataType(enum.Enum):
    # TODO - make capital
    GENOMIC_RANGES = "GenomicRanges"
    STR = "str"
    INT = "int"
    FLOAT = "float"
    ARRAY = "array"
    # TODO - add bed3 or bed6 ranges


@related.mutable(strict=False)
class MetadataStruct(RelatedConfigMixin):
    doc = related.StringField()
    type = related.ChildField(MetadataType, required=False)
    name = related.StringField(required=False)

    def compatible_with_batch(self, batch, verbose=True):
        """Checks compatibility with a particular numpy array

        Args:
          batch: numpy array of a batch

          verbose: print the fail reason
        """

        def print_msg(msg):
            if verbose:
                print("MetadataStruct mismatch")
                print(msg)

        # custom classess
        if self.type == MetadataType.GENOMIC_RANGES:
            if not isinstance(batch, GenomicRanges):
                # TODO - do we strictly require the GenomicRanges class?
                #          - relates to metadata.py TODO about numpy_collate
                #        for now we should just be able to convert to the GenomicRanges class
                #        without any errors
                try:
                    GenomicRanges.from_dict(batch)
                except Exception as e:
                    print_msg("expecting a GenomicRanges object or a GenomicRanges-like dict")
                    print_msg("convertion error: {0}".format(e))
                    return False
                else:
                    return True
            else:
                return True

        # type = np.ndarray
        if not isinstance(batch, np.ndarray):
            print_msg("Expecting a np.ndarray. Got type(batch) = {0}".format(type(batch)))
            return False

        if not batch.ndim >= 1:
            print_msg("The array is a scalar (expecting at least the batch dimension)")
            return False

        bshape = batch.shape[1:]

        # scalars
        if self.type in {MetadataType.INT, MetadataType.STR, MetadataType.FLOAT}:
            if bshape != () and bshape != (1,):
                print_msg("expecting a scalar, got an array with shape (without the batch axis): {0}".format(bshape))
                return False

        # arrays
        # - no checks

        return True


@related.mutable(strict=False)
class DataLoaderSchema(RelatedConfigMixin):
    """Describes the model schema

    Properties:
     - we allow classes that contain also dictionaries
       -> leaf can be an
         - array
         - scalar
         - custom dictionary (recursive data-type)
         - SpecialType (say ArrayRanges, BatchArrayRanges, which will
                        effectively be a dicitonary of scalars)
    """
    inputs = NestedMappingField(ArraySchema, keyword="shape", key="name")
    targets = NestedMappingField(ArraySchema, keyword="shape", key="name", required=False)
    metadata = NestedMappingField(MetadataStruct, keyword="doc", key="name",
                                  required=False)

    def compatible_with_batch(self, batch, verbose=True):
        """Validate if the batch of data complies with the schema

        Checks preformed:
        - nested structure is the same (i.e. dictionary names, list length etc)
        - array shapes are compatible
        - returned obj classess are compatible

        # Arguments
            batch: a batch of data returned by one iteraton of dataloader's batch_iter
                nested dictionary
            verbose: verbose error logging if things don't match

        # Returns
           bool: True only if everyhing is ok
        """

        def print_msg(msg):
            if verbose:
                print(msg)

        # check the individual names
        if not isinstance(batch, dict):
            print("not isinstance(batch, dict)")
            return False

        # contains only the three specified fields
        if not set(batch.keys()).issubset({"inputs", "targets", "metadata"}):
            print('not set(batch.keys()).issubset({"inputs", "targets", "metadata"})')
            return False

        # Inputs check
        def compatible_nestedmapping(batch, descr, cls, verbose=True):
            """Recursive function of checks

            shapes match, batch-dim matches
            """
            # we expect a numpy array/special class, a list or a dictionary

            # Special case for the metadat
            if isinstance(descr, cls):
                return descr.compatible_with_batch(batch, verbose=verbose)
            elif isinstance(batch, collections.Mapping) and isinstance(descr, collections.Mapping):
                if not set(batch.keys()) == set(descr.keys()):
                    print_msg("The dictionary keys don't match:")
                    print_msg("batch: {0}".format(batch.keys()))
                    print_msg("descr: {0}".format(descr.keys()))
                    return False
                return all([compatible_nestedmapping(batch[key], descr[key], cls, verbose) for key in batch])
            elif isinstance(batch, collections.Sequence) and isinstance(descr, collections.Sequence):
                if not len(batch) == len(descr):
                    print_msg("Lengths dont match:")
                    print_msg("len(batch): {0}".format(len(batch)))
                    print_msg("len(descr): {0}".format(len(descr)))
                    return False
                return all([compatible_nestedmapping(batch[i], descr[i], cls, verbose) for i in range(len(batch))])

            print_msg("Invalid types:")
            print_msg("type(batch): {0}".format(type(batch)))
            print_msg("type(descr): {0}".format(type(descr)))
            return False

        # inputs needs to be present allways
        if "inputs" not in batch:
            print_msg('not "inputs" in batch')
            return False

        if not compatible_nestedmapping(batch["inputs"], self.inputs, ArraySchema, verbose):
            return False

        if "targets" in batch and not \
                (len(batch["targets"]) == 0):  # unspecified
            if self.targets is None:
                # targets need to be specified if we want to use them
                print_msg('self.targets is None')
                return False
            if not compatible_nestedmapping(batch["targets"], self.targets, ArraySchema, verbose):
                return False

        # metadata needs to be present if it is defined in the description
        if self.metadata is not None:
            if "metadata" not in batch:
                print_msg('not "metadata" in batch')
                return False
            if not compatible_nestedmapping(batch["metadata"], self.metadata, MetadataStruct, verbose):
                return False
        else:
            if "metadata" in batch:
                print_msg('"metadata" in batch')
                return False

        return True


# --------------------------------------------
@related.mutable(strict=False)
class RemoteFile(RelatedConfigMixin):
    url = related.StringField()
    md5 = related.StringField("", required=False)
    name = related.StringField("", required=False)

    def __attrs_post_init__(self):
        if self.md5 == "":
            logger.warning("md5 not specified for url: {}".format(self.url))
        if os.path.basename(self.name) != self.name:
            logger.warning("'name' does not seem to be a valid file name: {}".format(self.name))
            self.name = os.path.basename(self.name)

    def validate(self, path):
        """Validate if the path complies with the provided md5 hash
        """
        return check_integrity(path, self.md5)

    def get_file(self, path):
        """Download the remote file to cache_dir and return
        the file path to it
        """
        if self.md5:
            file_hash = self.md5
        else:
            file_hash = None
        root, filename = os.path.dirname(path), os.path.basename(path)
        root = os.path.abspath(root)
        download_url(self.url, root, filename, file_hash)
        return os.path.join(root, filename)


@related.mutable(strict=False)
class DataLoaderArgument(RelatedConfigMixin):
    # MAYBE - make this a general argument class
    doc = related.StringField("", required=False)
    example = AnyField(required=False)
    default = AnyField(required=False)
    name = related.StringField(required=False)
    type = related.StringField(default='str', required=False)
    optional = related.BooleanField(default=False, required=False)
    tags = StrSequenceField(str, default=[], required=False)  # TODO - restrict the tags

    def __attrs_post_init__(self):
        if self.doc == "":
            logger.warning("doc empty for one of the dataloader `args` fields")
            # parse args
        self.example = recursive_dict_parse(self.example, 'url', RemoteFile.from_config)
        self.default = recursive_dict_parse(self.default, 'url', RemoteFile.from_config)


@related.mutable(strict=False)
class Dependencies(RelatedConfigMixin):
    conda = StrSequenceField(str, default=[], required=False, repr=True)
    pip = StrSequenceField(str, default=[], required=False, repr=True)
    # not really required
    conda_channels = related.SequenceField(str, default=["defaults"],
                                           required=False, repr=True)
    conda_file = related.StringField(required=False)

    def __attrs_post_init__(self):
        """
        In case conda or pip are filenames pointing to existing files,
        read the files and populate the package names
        """
        if self.conda_file:
            # use the dependencies from the conda file. Override conda, pip and conda_file
            deps = read_yaml(self.conda_file)
            pip_deps = [x['pip'] for x in deps['dependencies']
                        if isinstance(x, dict)][0]
            conda_deps = [x for x in deps['dependencies']
                          if not isinstance(x, dict)]
            object.__setattr__(self, "pip", pip_deps)
            object.__setattr__(self, "conda", conda_deps)
            object.__setattr__(self, "conda_channels", deps['channels'])
        if len(self.conda) == 1 and self.conda[0].endswith(".txt") and \
                os.path.exists(self.conda[0]):
            # found a conda txt file
            object.__setattr__(self, "conda", read_txt(self.conda[0]))

        if len(self.pip) == 1 and self.pip[0].endswith(".txt") and \
                os.path.exists(self.pip[0]):
            # found a pip txt file
            object.__setattr__(self, "pip", read_txt(self.pip[0]))

    def all_installed(self, verbose=False):
        """Validate if all the dependencies are installed as requested

        Args:
          verbose: if True, display warnings if the dependencies are not installed

        Returns:
          (bool): True if all the required package versions are installed
            and False otherwise
        """
        norm = self.normalized()
        for pkg in list(norm.conda) + list(norm.pip):
            if not kconda.is_installed(pkg):
                if verbose:
                    pkg_name, req_version = kconda.version_split(pkg)
                    found_version = kconda.get_package_version(pkg_name)
                    if found_version is None:
                        print("Package '{}' is not installed".
                              format(pkg_name))
                    else:
                        print("Installed package '{}={}' doesn't "
                              "comply with '{}'".
                              format(pkg_name, found_version, pkg))
                return False
        return True

    def install_pip(self, dry_run=False):
        print("pip dependencies to be installed:")
        print(self.pip)
        if dry_run:
            return
        else:
            kconda.install_pip(self.pip)

    def install_conda(self, dry_run=False):
        print("Conda dependencies to be installed:")
        print(self.conda)
        if dry_run:
            return
        else:
            channels, packages = self._get_channels_packages()
            kconda.install_conda(packages, channels)

    def install(self, dry_run=False):
        self.install_conda(dry_run)
        self.install_pip(dry_run)

    def merge(self, dependencies):
        """Merge one dependencies with another one

        Use case: merging the dependencies of model and dataloader

        Args:
            dependencies: Dependencies instance

        Returns:
            new Dependencies instance
        """
        return Dependencies(
            conda=unique_list(list(self.conda) + list(dependencies.conda)),
            pip=kconda.normalize_pip(list(self.pip) + list(dependencies.pip)),
            conda_channels=unique_list(list(self.conda_channels) + list(dependencies.conda_channels))
        )

    def normalized(self):
        """Normalize the list of dependencies
        """
        channels, packages = self._get_channels_packages()
        if isinstance(packages, related.types.TypedSequence):
            packages = packages.list
        if isinstance(channels, related.types.TypedSequence):
            channels = channels.list

        return Dependencies(
            conda=packages,
            pip=kconda.normalize_pip(list(self.pip)),
            conda_channels=channels)

    def _get_channels_packages(self):
        """Get conda channels and packages separated from each other(by '::')
        """
        if len(self.conda) == 0:
            return self.conda_channels, self.conda
        channels, packages = list(zip(*map(kconda.parse_conda_package, self.conda)))
        channels = unique_list(list(channels) + list(self.conda_channels))
        packages = unique_list(list(packages))

        # Handle channel order
        if "bioconda" in channels and "conda-forge" not in channels:
            # Insert 'conda-forge' right after bioconda if it is not included
            channels.insert(channels.index("bioconda") + 1, "conda-forge")
        if "pysam" in packages and "bioconda" in channels:
            if channels.index("defaults") < channels.index("bioconda"):
                logger.warning("Swapping channel order - putting defaults last. " +
                               "Using pysam bioconda instead of anaconda")
                channels.remove("defaults")
                channels.insert(len(channels), "defaults")
        return channels, packages

    def to_env_dict(self, env_name):
        deps = self.normalized()
        channels, packages = deps._get_channels_packages()
        if isinstance(packages, related.types.TypedSequence):
            packages = packages.list
        if isinstance(channels, related.types.TypedSequence):
            channels = channels.list

        env_dict = OrderedDict(
            name=env_name,
            channels=channels,
            dependencies=packages + [OrderedDict(pip=kconda.normalize_pip(deps.pip))]
        )
        return env_dict

    @classmethod
    def from_env_dict(self, dict):
        cfg = {}
        cfg["conda_channels"] = dict['channels']
        cfg["conda"] = [el for el in dict['dependencies'] if not isinstance(el, OrderedDict)]
        pip = [el for el in dict['dependencies'] if isinstance(el, OrderedDict)]
        if len(pip) == 1:
            cfg["pip"] = pip[0]['pip']
        elif len(pip) > 1:
            raise Exception("Malformatted conda environment yaml!")
        return self.from_config(cfg)

    def to_env_file(self, env_name, path):
        """Dump the dependencies to a file
        """
        with open(path, 'w') as f:
            d = self.to_env_dict(env_name)

            # add python if not present
            add_py = True
            for dep in d['dependencies']:
                if isinstance(dep, str) and dep.startswith("python"):
                    add_py = False

            if add_py:
                d['dependencies'] = ["python"] + d['dependencies']
            # -----
            # remove fields that are empty
            out = []
            for k in d:
                if not (isinstance(d[k], list) and len(d[k]) == 0):
                    out.append((k, d[k]))
            # -----

            f.write(yaml_ordered_dump(OrderedDict(out),
                                      indent=2,
                                      default_flow_style=False))

    def gpu(self):
        """Get the gpu - version of the dependencies
        """

        def replace_gpu(dep):
            if dep.startswith("tensorflow") and "gpu" not in dep:
                new_dep = dep.replace("tensorflow", "tensorflow-gpu")
                logger.info("use gpu: Replacing the dependency {0} with {1}".format(dep, new_dep))
                return new_dep
            if dep.startswith("pytorch-cpu"):
                new_dep = dep.replace("pytorch-cpu", "pytorch")
                logger.info("use gpu: Replacing the dependency {0} with {1}".format(dep, new_dep))
                return new_dep
            return dep

        deps = self.normalized()
        return Dependencies(
            conda=[replace_gpu(dep) for dep in deps.conda],
            pip=[replace_gpu(dep) for dep in deps.pip],
            conda_channels=deps.conda_channels)

    def osx(self):
        """Get the os - x compatible dependencies
        """
        from sys import platform
        if platform != 'darwin':
            logger.warning("Calling osx dependency conversion on non-osx platform: {}".
                           format(platform))

        def replace_osx(dep):
            if dep.startswith("pytorch-cpu"):
                new_dep = dep.replace("pytorch-cpu", "pytorch")
                logger.info("osx: Replacing the dependency {0} with {1}".
                            format(dep, new_dep))
                return new_dep
            return dep

        deps = self.normalized()
        return Dependencies(
            conda=[replace_osx(dep) for dep in deps.conda],
            pip=[replace_osx(dep) for dep in deps.pip],
            conda_channels=deps.conda_channels)

    # @classmethod
    # def from_file(cls, path):
    #     """TODO instantiate Dependencies from a yaml file
    #     """
    #     pass


@related.mutable(strict=False)
class DataLoaderImport(RelatedConfigMixin):
    """Dataloader specification for the import
    """
    defined_as = related.StringField()
    default_args = related.ChildField(dict, default=OrderedDict(), required=False)
    # specify also dataloader dependencies explicitly
    dependencies = related.ChildField(Dependencies,
                                      default=Dependencies(),
                                      required=False)
    # whether to parse the dependencies from the dataloader when installing it
    parse_dependencies = related.BooleanField(default=True, required=False)

    def get(self):
        """Get the dataloader
        """
        from kipoi.data import BaseDataLoader
        from copy import deepcopy
        obj = load_obj(self.defined_as)

        # check that it inherits from BaseDataLoader
        if not inherits_from(obj, BaseDataLoader):
            raise ValueError("Dataloader: {} doen't inherit from kipoi.data.BaseDataLoader".format(self.defined_as))

        # override the default arguments
        if self.default_args:
            obj = override_default_kwargs(obj, self.default_args)

        # override also the values in the example in case
        # they were previously specified
        for k, v in six.iteritems(self.default_args):
            if not isinstance(obj.args[k].example, UNSPECIFIED):
                obj.args[k].example = v

        return obj


@related.mutable(strict=False)
class ModelTest(RelatedLoadSaveMixin):
    # predictions = related.
    expect = AnyField(default=None, required=False)
    precision_decimal = related.IntegerField(default=7, required=False)

    # Arrays should be almost equal to `precision_decimal` places
    # https://docs.scipy.org/doc/numpy-1.15.1/reference/generated/numpy.testing.assert_almost_equal.html
    # abs(desired-actual) < 1.5 * 10**(-precision_decimal)

    def __attrs_post_init__(self):
        if self.expect is not None:
            if not isinstance(self.expect, str):
                # it has to be the url
                if not (isinstance(self.expect, dict) and "url" in self.expect):
                    raise ValueError("expect is not a file path, expecting a url field with entries: url and md5")
                self.expect = RemoteFile.from_config(self.expect)


# --------------------------------------------
# Final description classes modelling the yaml files
@related.mutable(strict=False)
class ModelDescription(RelatedLoadSaveMixin):
    """Class representation of model.yaml
    """
    args = related.ChildField(dict)
    info = related.ChildField(ModelInfo)
    schema = related.ChildField(ModelSchema)
    defined_as = related.StringField(required=False)
    type = related.StringField(required=False)
    default_dataloader = AnyField(default='.', required=False)
    postprocessing = related.ChildField(dict, default=OrderedDict(), required=False)
    dependencies = related.ChildField(Dependencies,
                                      default=Dependencies(),
                                      required=False)
    test = related.ChildField(ModelTest,
                              default=ModelTest(),
                              required=False)
    path = related.StringField(required=False)

    # TODO - add after loading validation for the arguments class?

    def __attrs_post_init__(self):
        if self.defined_as is None and self.type is None:
            raise ValueError("Either defined_as or type need to be specified")
        # load additional objects
        for k in self.postprocessing:
            k_observed = k
            if k == 'variant_effects':
                k = 'kipoi_veff'
            if is_installed(k):
                # Load the config properly if the plugin is installed
                try:
                    parser = get_model_yaml_parser(k)
                    self.postprocessing[k_observed] = parser.from_config(self.postprocessing[k_observed])

                    object.__setattr__(self, "postprocessing", self.postprocessing)
                except Exception:
                    logger.warning("Unable to parse {} filed in ModelDescription: {}".format(k_observed, self))

        # parse args
        self.args = recursive_dict_parse(self.args, 'url', RemoteFile.from_config)

        # parse default_dataloader
        if isinstance(self.default_dataloader, dict):
            self.default_dataloader = DataLoaderImport.from_config(self.default_dataloader)


def example_kwargs(dl_args, cache_path=None, absolute_path=True, dry_run=False):
    """Return the example kwargs.

    Args:
      dl_args: dictionary of dataloader args
      cache_path: if specified, save the examples to that directory
    """
    example_files = {}
    for k, v in six.iteritems(dl_args):
        if isinstance(v.example, UNSPECIFIED):
            continue
        if isinstance(v.example, RemoteFile) and cache_path is not None:
            if absolute_path:
                dl_dir = os.path.abspath(cache_path)
            else:
                dl_dir = cache_path
            if not os.path.exists(dl_dir):
                os.makedirs(dl_dir)

            # determine target path of the example file
            if v.example.name != "":
                # use file name as provided in the example
                path = os.path.join(dl_dir, v.example.name)
            else:
                # otherwise just call it like the argument
                path = os.path.join(dl_dir, k)

            example_files[k] = path
            if os.path.exists(path):
                if v.example.validate(path):
                    logger.info("Example file for argument {} already exists".format(k))
                else:
                    logger.info("Example file for argument {} doesn't match the md5 "
                                "hash {}. Re-downloading".format(k, v.example.md5))
                    if not dry_run:
                        v.example.get_file(path)  # TODO
            else:
                if not dry_run:
                    v.example.get_file(path)  # TODO
        else:
            example_files[k] = v.example
    return example_files
    # return {k: v.example for k, v in six.iteritems(dl_args) if not isinstance(v.example, UNSPECIFIED)}


def download_default_args(args, output_dir):
    """Download the default files
    """
    override = {}
    for k in args:
        # arg.default is None
        if args[k].default is not None:
            if isinstance(args[k].default, UNSPECIFIED):
                continue
            if isinstance(args[k].default, RemoteFile):
                # if it's a remote file, download it
                # and set the default to the file path

                # specify the file name and create the directory
                logger.info("Downloading dataloader default arguments {} from {}".format(k, args[k].default.url))
                if not os.path.exists(output_dir):
                    os.makedirs(output_dir)
                if args[k].default.md5:
                    fname = args[k].default.md5
                else:
                    fname = "file"

                # download the parameters and override the model
                args[k].default = args[k].default.get_file(os.path.join(output_dir, fname))

            # build up a override dict from .default args
            if isinstance(args[k].default, bool):
                # necessary check for booleans
                override[k] = args[k].default
            elif os.path.exists(args[k].default):
                # for files, make sure we are using absolute paths
                override[k] = os.path.abspath(args[k].default)
            else:
                override[k] = args[k].default
    return override


@related.mutable(strict=False)
class DataLoaderDescription(RelatedLoadSaveMixin):
    """Class representation of dataloader.yaml
    """
    defined_as = related.StringField()
    args = related.MappingField(DataLoaderArgument, "name")
    output_schema = related.ChildField(DataLoaderSchema)
    type = related.StringField(required=False)
    info = related.ChildField(Info, default=Info(), required=False)
    dependencies = related.ChildField(Dependencies, default=Dependencies(), required=False)
    path = related.StringField(required=False)
    postprocessing = related.ChildField(dict, default=OrderedDict(), required=False)

    def get_example_kwargs(self):
        # return self.download_example()
        if self.path is None:
            path = "."
        else:
            path = self.path
        return example_kwargs(self.args, os.path.join(os.path.dirname(path), "downloaded/example_files"))

    def download_example(self, output_dir, absolute_path=False, dry_run=False):
        return example_kwargs(self.args,
                              output_dir,
                              absolute_path=absolute_path,
                              dry_run=dry_run)

    def print_kwargs(self, format_examples_json=False):
        from kipoi_utils.external.related.fields import UNSPECIFIED
        if not hasattr(self, "args"):
            logger.warning("No keyword arguments defined for the given dataloader.")
            return None

        args = self.args
        for k in args:
            print("{0}:".format(k))
            for elm in ["doc", "type", "optional", "example"]:
                if hasattr(args[k], elm) and \
                        (not isinstance(getattr(args[k], elm), UNSPECIFIED)):
                    print("    {0}: {1}".format(elm, getattr(args[k], elm)))

        # example_kwargs = self.get_example_kwargs()
        # if format_examples_json:
        #     import json
        #     example_kwargs = json.dumps(example_kwargs)
        # print("Example keyword arguments are: {0}".format(str(example_kwargs)))

    print_args = print_kwargs

    def __attrs_post_init__(self):
        # load additional objects
        for k in self.postprocessing:
            k_observed = k
            if k == 'variant_effects':
                k = 'kipoi_veff'
            if is_installed(k):
                # Load the config properly if the plugin is installed
                try:
                    parser = get_dataloader_yaml_parser(k)
                    self.postprocessing[k_observed] = parser.from_config(self.postprocessing[k_observed])
                    object.__setattr__(self, "postprocessing", self.postprocessing)
                except Exception:
                    logger.warning("Unable to parse {} filed in DataLoaderDescription: {}".format(k_observed, self))

    # download example files
    # def download_example(self):
    #     example_files = {}
    #     for k, v in six.iteritems(self.args):
    #         if isinstance(v.example, RemoteFile):
    #             if self.path is None:
    #                 raise ValueError("Unable to download example files. path attribute not specified")

    #             dl_dir = os.path.join(self.path, "dataloader_files")
    #             if not os.path.exists(dl_dir):
    #                 os.makedirs(dl_dir)
    #             path = os.path.join(dl_dir, k)
    #             example_files[k] = path
    #             if os.path.exists(path):
    #                 if v.example.validate(path):
    #                     logger.info("Example file for argument {} already exists".format(k))
    #                 else:
    #                     logger.info("Example file for argument {} doesn't match the md5 hash {}. Re-downloading".format(k))
    #                     v.example.get_file(path)  # TODO
    #             else:
    #                 v.example.get_file(path)  # TODO
    #         else:
    #             example_files[k] = v
    #     return example_files


# ---------------------
# Global source config


# TODO - write a unit-test for these three
@related.mutable
class TestModelConfig(RelatedConfigMixin):
    batch_size = related.IntegerField(default=None, required=False)


@related.mutable
class TestConfig(RelatedConfigMixin):
    """Models config.yaml in the model root
    """
    constraints = related.MappingField(TestModelConfig, "name", required=False,
                                       repr=True)


@related.mutable
class SourceConfig(RelatedLoadSaveMixin):
    test = related.ChildField(TestConfig, required=False)
    # default dependencies
    dependencies = related.ChildField(Dependencies,
                                      default=Dependencies(),
                                      required=False)
    path = related.StringField(required=False)

# TODO - special metadata classes should just extend the dictionary field
# (to be fully compatible with batching etc)



from __future__ import absolute_import
from __future__ import print_function

import sys
import re
import os
import abc
import six
import textwrap
import inspect
from collections import OrderedDict

import related
import kipoi  # for .config module
from kipoi.specs import DataLoaderDescription, Info, example_kwargs, RemoteFile, download_default_args
from kipoi_utils import (load_module, cd, getargs, classproperty, inherits_from, rsetattr,
                         _get_arg_name_values, load_obj, infer_parent_class, override_default_kwargs)
from kipoi_utils.external.torch.data import DataLoader
from kipoi_utils.data_utils import (numpy_collate, numpy_collate_concat, get_dataset_item,
                                    DataloaderIterable, batch_gen, get_dataset_lens, iterable_cycle)
from tqdm import tqdm
import types

import logging

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

#
PREPROC_IFILE_TYPES = ['DNA_regions']
PREPROC_IFILE_FORMATS = ['bed3']


class BaseDataLoader(object):
    """Abstract Dataloader class
    """
    __metaclass__ = abc.ABCMeta

    # dataloader descriptors. Parsed from DataLoaderSchema in _add_description_factory()
    type = None
    defined_as = None
    args = None
    output_schema = None
    info = None
    dependencies = None
    path = None
    postprocessing = None
    # optionally set in get_dataloader_factory
    source = None
    source_dir = None

    # ---------------------------------
    # Core data-loading methods

    @abc.abstractmethod
    def batch_iter(self, **kwargs):
        raise NotImplementedError

    def batch_train_iter(self, cycle=True, **kwargs):
        """Returns samples directly useful for training the model:
        (x["inputs"],x["targets"])

        Args:
          cycle: when True, the returned iterator will run indefinitely go through the dataset
            Use True with `fit_generator` in Keras.
          **kwargs: Arguments passed to self.batch_iter(**kwargs)
        """
        if cycle:
            return ((x["inputs"], x["targets"])
                    for x in iterable_cycle(self._batch_iterable(**kwargs)))
        else:
            return ((x["inputs"], x["targets"]) for x in self.batch_iter(**kwargs))

    def batch_predict_iter(self, **kwargs):
        """Returns samples directly useful for prediction x["inputs"]

        Args:
          **kwargs: Arguments passed to self.batch_iter(**kwargs)
        """
        return (x["inputs"] for x in self.batch_iter(**kwargs))

    def load_all(self, **kwargs):
        """Loads and returns the whole dataset

        Arguments:
            **kwargs: passed to batch_iter()
        """
        return numpy_collate_concat([x for x in tqdm(self.batch_iter(**kwargs))])

    # ---------------------------------
    # Book-keeping methods dealing with
    # the annotation of the dataloader
    @classmethod
    def _add_description_factory(cls, descr):
        """Factory method which populates the un-set class variables

        Returns:
          new dataloader class
        """
        for field in ['type', 'defined_as', 'args', 'output_schema',
                      'info', 'dependencies', 'path', 'postprocessing']:
            setattr(cls, field, getattr(descr, field))
        return cls

    @classproperty
    def example_kwargs(cls):
        if cls.args is None:
            raise ValueError("Class description `args` is missing. "
                             "Use `_add_description_factory` to annotate the class")
        if cls.source_dir is None:
            logger.info("Using current directory for source_dir")
            cls.source_dir = os.getcwd()

        # Add init_example method.
        # example_kwargs also downloads files to {dataloader_dir}/dataloader_files
        return example_kwargs(cls.args, os.path.join(cls.source_dir, "downloaded/example_files"))

    @classmethod
    def download_example(cls, output_dir, absolute_path=False, dry_run=False):
        """Download the example files to the desired directory

        # Arguments
          output_dir: output directory where to store the file
          absolute_path: if True, return absolute paths to the
            output directories
          dry_run: if True, return only the file paths without
            actually downloading the files

        # Returns
          dictionary of keyword arguments for the dataloader
        """
        return example_kwargs(cls.args, output_dir, absolute_path=absolute_path, dry_run=dry_run)

    @classmethod
    def init_example(cls):
        """Instantiate the class using example_kwargs
        """
        if cls.source_dir is not None:
            with cd(cls.source_dir):
                # always init the example in the original directory
                return cls(**cls.example_kwargs)
        else:
            return cls(**cls.example_kwargs)

    @classmethod
    def print_args(cls, format_examples_json=False):
        """Print dataloader kwargs

        # Arguments
          format_examples_json: format the results as json
        """
        from kipoi_utils.external.related.fields import UNSPECIFIED
        if not hasattr(cls, "args"):
            logger.warning("No keyword arguments defined for the given dataloader.")
            return None
            # print("No keyword arguments defined for the given dataloader.")
        args = cls.args
        for k in args:
            print("{0}:".format(k))
            for elm in ["doc", "type", "optional", "example"]:
                if hasattr(args[k], elm) and \
                        (not isinstance(getattr(args[k], elm), UNSPECIFIED)):
                    print("    {0}: {1}".format(elm, getattr(args[k], elm)))
        # example_kwargs = cls.example_kwargs
        # print("-" * 80)
        # if hasattr(cls, "example_kwargs"):
        #     if format_examples_json:
        #         import json
        #         example_kwargs = json.dumps(example_kwargs)
        #     print("Example keyword arguments are: {0}".format(str(example_kwargs)))

    @classmethod
    def get_output_schema(cls):
        return cls.output_schema


def kipoi_dataloader(override=dict()):
    """Decorator for converting a Dataloader class with dataloader.yaml description in the docstring
    into a proper Kipoi dataloader

    It parses the doc-string of the class as the DataloaderDescription
    and populates the class description attributes with it

    # Arguments
        override: dictionary containing values to override or specify in the decorated class.
          supports nesting the attributes. example: `{'info.authors': [Author(name='name')]}`

    # __call__
        dataloader containing the descripiton specified in the yaml doc-string
    """

    def wrap(cls):
        if inspect.isfunction(cls):
            raise ValueError("Function-based dataloader are currently not supported with kipoi_dataloader decorator")

        # figure out the right dataloader type
        dl_type_inferred = infer_parent_class(cls, AVAILABLE_DATALOADERS)
        if dl_type_inferred is None:
            raise ValueError("Dataloader needs to inherit from one of the available dataloaders {}".format(
                list(AVAILABLE_DATALOADERS)))

        # or not inherits_from(cls, Dataset)
        doc = cls.__doc__
        doc = textwrap.dedent(doc)  # de-indent

        if not re.match("^defined_as: ", doc):
            doc = "defined_as: {}\n".format(cls.__name__) + doc
        if not re.match("^type: ", doc):
            doc = "type: {}\n".format(dl_type_inferred) + doc

        # parse the yaml
        yaml_dict = related.from_yaml(doc)
        dl_descr = DataLoaderDescription.from_config(yaml_dict)

        # override parameters
        for k, v in six.iteritems(override):
            rsetattr(dl_descr, k, v)

        # setup optional parameters
        arg_names, default_values = _get_arg_name_values(cls)

        if set(dl_descr.args) != set(arg_names):
            raise ValueError("Described args don't exactly match the implemented arguments"
                             "docstring: {}, actual: {}".format(list(dl_descr.args), list(arg_names)))

        # properly set optional / non-optional argument values
        for i, arg in enumerate(dl_descr.args):
            optional = i >= len(arg_names) - len(default_values)
            if dl_descr.args[arg].optional and not optional:
                logger.warning("Parameter {} was specified as optional. However, there "
                               "are no defaults for it. Specifying it as not optinal".format(arg))
            dl_descr.args[arg].optional = optional

        dl_descr.info.name = cls.__name__

        # enrich the class with dataloader description
        return cls._add_description_factory(dl_descr)

    return wrap


# --------------------------------------------
# Different implementations

# Other options:
# - generator - sample, batch-based
#   - yield.
# - iterator, iterable - sample, batch-based
#   - __iter__()
#     - __next__()
# - full dataset
#   - everything numpy arrays with the same first axis length


class PreloadedDataset(BaseDataLoader):
    """Generated by supplying a function returning the full dataset.

    The full dataset is a nested (list/dict) python structure of numpy arrays
    with the same first axis dimension.
    """
    data_fn = None

    @classmethod
    def from_fn(cls, data_fn):
        """setup the class variable
        """
        cls.data_fn = staticmethod(data_fn)
        return cls

    @classmethod
    def from_data(cls, data):
        return cls.from_data_fn(lambda: data)()

    @classmethod
    def _get_data_fn(cls):
        assert cls.data_fn is not None
        return cls.data_fn

    def __init__(self, *args, **kwargs):
        self.data = self._get_data_fn()(*args, **kwargs)
        lens = get_dataset_lens(self.data, require_numpy=True)
        # check that all dimensions are the same
        assert len(set(lens)) == 1
        self.n = lens[0]

    def __len__(self):
        return self.n

    def __getitem__(self, index):
        return get_dataset_item(self.data, index)

    def _batch_iterable(self, batch_size=32, shuffle=False, drop_last=False, **kwargs):
        """See batch_iter docs

        Returns:
          iterable
        """
        dl = DataLoader(self, batch_size=batch_size,
                        collate_fn=numpy_collate,
                        shuffle=shuffle,
                        num_workers=0,
                        drop_last=drop_last)
        return dl

    def batch_iter(self, batch_size=32, shuffle=False, drop_last=False, **kwargs):
        """Return a batch-iterator

        Arguments:
            dataset (Dataset): dataset from which to load the data.
            batch_size (int, optional): how many samples per batch to load
                (default: 1).
            shuffle (bool, optional): set to ``True`` to have the data reshuffled
                at every epoch (default: False).
            drop_last (bool, optional): set to ``True`` to drop the last incomplete batch,
                if the dataset size is not divisible by the batch size. If False and
                the size of dataset is not divisible by the batch size, then the last batch
                will be smaller. (default: False)

        Returns:
            iterator
        """
        dl = self._batch_iterable(batch_size=batch_size,
                                  shuffle=shuffle,
                                  drop_last=drop_last,
                                  **kwargs)
        return iter(dl)

    def load_all(self, **kwargs):
        """Load the whole dataset into memory

        Arguments:
            **kwargs: ignored
        """
        return self.data


class Dataset(BaseDataLoader):
    """An abstract class representing a Dataset.

    All other datasets should subclass it. All subclasses should override
    ``__len__``, that provides the size of the dataset, and ``__getitem__``,
    supporting integer indexing in range from 0 to len(self) exclusive.
    """

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def __getitem__(self, index):
        """Return one sample

        index: {0, ..., len(self)-1}
        """
        raise NotImplementedError

    @abc.abstractmethod
    def __len__(self):
        """Return the number of all samples
        """
        raise NotImplementedError

    def _batch_iterable(self, batch_size=32, shuffle=False, num_workers=0, drop_last=False, **kwargs):
        """Return a batch-iteratrable

        See batch_iter docs

        Returns:
            Iterable
        """
        dl = DataLoader(self,
                        batch_size=batch_size,
                        collate_fn=numpy_collate,
                        shuffle=shuffle,
                        num_workers=num_workers,
                        drop_last=drop_last,
                        **kwargs)
        return dl

    def batch_iter(self, batch_size=32, shuffle=False, num_workers=0, drop_last=False, **kwargs):
        """Return a batch-iterator

        Arguments:
            dataset (Dataset): dataset from which to load the data.
            batch_size (int, optional): how many samples per batch to load
                (default: 1).
            shuffle (bool, optional): set to ``True`` to have the data reshuffled
                at every epoch (default: False).
            num_workers (int, optional): how many subprocesses to use for data
                loading. 0 means that the data will be loaded in the main process
                (default: 0)
            drop_last (bool, optional): set to ``True`` to drop the last incomplete batch,
                if the dataset size is not divisible by the batch size. If False and
                the size of dataset is not divisible by the batch size, then the last batch
                will be smaller. (default: False)

        Returns:
            iterator
        """
        dl = self._batch_iterable(batch_size=batch_size,
                                  shuffle=shuffle,
                                  num_workers=num_workers,
                                  drop_last=drop_last,
                                  **kwargs)
        return iter(dl)

    def load_all(self, batch_size=32, **kwargs):
        """Load the whole dataset into memory
        Arguments:
            batch_size (int, optional): how many samples per batch to load
                (default: 1).
        """
        return numpy_collate_concat([x for x in tqdm(self.batch_iter(batch_size, **kwargs))])


class BatchDataset(BaseDataLoader):
    """An abstract class representing a BatchDataset.
    """

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def __getitem__(self, index):
        """Return one batch
        """
        raise NotImplementedError

    @abc.abstractmethod
    def __len__(self):
        """Number of all batches
        """
        raise NotImplementedError

    def _batch_iterable(self, num_workers=0, **kwargs):
        """Return a batch-iteratorable

        See batch_iter for docs
        """
        dl = DataLoader(self, batch_size=1,
                        collate_fn=numpy_collate_concat,
                        shuffle=False,
                        num_workers=num_workers,
                        drop_last=False)
        return dl

    def batch_iter(self, num_workers=0, **kwargs):
        """Return a batch-iterator

        Arguments:
            dataset (Dataset): dataset from which to load the data.
            num_workers (int, optional): how many subprocesses to use for data
                loading. 0 means that the data will be loaded in the main process
                (default: 0)
        Returns:
            iterator
        """
        dl = self._batch_iterable(num_workers=num_workers, **kwargs)
        return iter(dl)


class SampleIterator(BaseDataLoader):
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def __iter__(self):
        raise NotImplementedError

    # TODO - how to maintain compatibility with python2?
    @abc.abstractmethod
    def __next__(self):
        raise NotImplementedError

    next = __next__

    def batch_iter(self, batch_size=32, **kwargs):
        return batch_gen(iter(self), batch_size=batch_size)

    def _batch_iterable(self, batch_size=32, **kwargs):
        kwargs['batch_size'] = batch_size
        return DataloaderIterable(self, kwargs)


class BatchIterator(BaseDataLoader):

    @abc.abstractmethod
    def __iter__(self):
        raise NotImplementedError

    # TODO - how to maintain compatibility with python2?
    @abc.abstractmethod
    def __next__(self):
        raise NotImplementedError

    next = __next__

    def batch_iter(self, **kwargs):
        return iter(self)

    def _batch_iterable(self, **kwargs):
        return DataloaderIterable(self, kwargs)


class SampleGenerator(BaseDataLoader):
    """Transform a generator of samples into SampleIterator
    """
    generator_fn = None

    @classmethod
    def from_fn(cls, generator_fn):
        """setup the class variable
        """
        cls.generator_fn = staticmethod(generator_fn)
        return cls

    @classmethod
    def _get_generator_fn(cls):
        assert cls.generator_fn is not None
        return cls.generator_fn

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __iter__(self):
        """Return a new generator every time
        """
        return self._get_generator_fn()(*self.args, **self.kwargs)

    def batch_iter(self, batch_size=32, **kwargs):
        return batch_gen(iter(self), batch_size=batch_size)

    def _batch_iterable(self, batch_size=32, **kwargs):
        kwargs['batch_size'] = batch_size
        return DataloaderIterable(self, kwargs)


class BatchGenerator(BaseDataLoader):
    """Transform a generator of batches into BatchIterator
    """
    generator_fn = None

    @classmethod
    def from_fn(cls, generator_fn):
        cls.generator_fn = staticmethod(generator_fn)
        return cls

    @classmethod
    def _get_generator_fn(cls):
        assert cls.generator_fn is not None
        return cls.generator_fn

    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs

    def __iter__(self):
        return self._get_generator_fn()(*self.args, **self.kwargs)

    def batch_iter(self, **kwargs):
        return iter(self)

    def _batch_iterable(self, **kwargs):
        return DataloaderIterable(self, kwargs)


# --------------------------------------------


def get_dataloader(dataloader, source="kipoi"):
    """Loads the dataloader

    # Arguments
        dataloader (str): dataloader name
        source (str): source name

    # Returns
    - Instance of class inheriting from `kipoi.data.BaseDataLoader` (like `kipoi.data.Dataset`)
           decorated with additional attributes.

    # Methods
    - __batch_iter(batch_size, num_workers, **kwargs)__
         - Arguments
             - **batch_size**: batch size
             - **num_workers**: Number of workers to use in parallel.
             - ****kwargs**: Other kwargs specific to each dataloader
         - Yields
             - `dict` with `"inputs"`, `"targets"` and `"metadata"`
    - __batch_train_iter(cycle=True, **kwargs)__
         - Arguments
             - **cycle**: if True, cycle indefinitely
             - ****kwargs**: Kwargs passed to `batch_iter()` like `batch_size`
         - Yields
             - tuple of ("inputs", "targets") from the usual dict returned by `batch_iter()`
    - __batch_predict_iter(**kwargs)__
         - Arguments
             - ****kwargs**: Kwargs passed to `batch_iter()` like `batch_size`
         - Yields
             - "inputs" field from the usual dict returned by `batch_iter()`
    - __load_all(**kwargs)__ - load the whole dataset into memory
         - Arguments
             - ****kwargs**: Kwargs passed to `batch_iter()` like `batch_size`
         - Returns
             - `dict` with `"inputs"`, `"targets"` and `"metadata"`
    - **init_example()** - instantiate the dataloader with example kwargs
    - **print_args()** - print information about the required arguments

    # Appended attributes
    - **type** (str): dataloader type (class name)
    - **defined_as** (str): path and dataloader name
    - **args** (list of kipoi.specs.DataLoaderArgument): datalaoder argument description
    - **info** (kipoi.specs.Info): general information about the dataloader
    - **schema** (kipoi.specs.DataloaderSchema): information about the input/output
            data modalities
    - **dependencies** (kipoi.specs.Dependencies): class specifying the dependencies.
          (implements `install` method for running the installation)
    - **name** (str): model name
    - **source** (str): model source
    - **source_dir** (str): local path to model source storage
    - **postprocessing** (dict): dictionary of loaded plugin specifications
    - **example_kwargs** (dict): kwargs for running the provided example
    """
    # if source == 'py':
    #     # load it from the python object
    #     sys.path.append(os.path.getcwd())
    #     return DataLoaderImport(defined_as=dataloader).get()
    # TODO - allow source=py

    # pull the dataloader & get the dataloader directory
    if isinstance(source, str):
        source = kipoi.config.get_source(source)
    source.pull_dataloader(dataloader)
    dataloader_dir = source.get_dataloader_dir(dataloader)

    # --------------------------------------------
    # Setup dataloader description
    descr = source.get_dataloader_descr(dataloader)
    with cd(dataloader_dir):  # move to the dataloader directory temporarily
        if "::" in descr.defined_as:
            # old API
            file_path, obj_name = tuple(descr.defined_as.split("::"))
            CustomDataLoader = getattr(load_module(file_path), obj_name)
        else:
            # new API - directly specify the object
            CustomDataLoader = load_obj(descr.defined_as)

    # download util links if specified under default & override the default parameters
    override = download_default_args(descr.args, source.get_dataloader_download_dir(dataloader))
    if override:
        # override default arguments specified under default
        CustomDataLoader = override_default_kwargs(CustomDataLoader, override)

    # infer the type
    if descr.type is None:
        if inspect.isfunction(CustomDataLoader):
            raise ValueError(
                "Dataloaders implemented as functions/generator need to specify the type flag in dataloader.yaml")
        else:
            # figure out the right dataloader type
            descr.type = infer_parent_class(CustomDataLoader, AVAILABLE_DATALOADERS)
            if descr.type is None:
                raise ValueError("Dataloader needs to inherit from one of the available dataloaders {}".format(
                    list(AVAILABLE_DATALOADERS)))

        # check that descr.type is correct
    if descr.type not in AVAILABLE_DATALOADERS:
        raise ValueError("dataloader type: {0} is not in supported dataloaders:{1}".
                         format(descr.type, list(AVAILABLE_DATALOADERS.keys())))

    # check that the extractor arguments match yaml arguments
    if not getargs(CustomDataLoader) == set(descr.args.keys()):
        # if kwargs is defined, we can assume that the dataloader accepts more than its explicitly defined arguments
        if "kwargs" in getargs(CustomDataLoader):
            missing_required_args = getargs(CustomDataLoader) - set(descr.args.keys()) - {"kwargs"}
            if len(missing_required_args) > 0:
                raise ValueError(
                    "DataLoader arguments:\n\t{args}\n are missing required arguments:\n\t{missing}".format(
                        args=set(descr.args.keys()),
                        missing=missing_required_args,
                    )
                )
        else:
            raise ValueError(
                "DataLoader arguments: \n{0}\n don't match ".format(set(getargs(CustomDataLoader))) +
                "the specification in the dataloader.yaml file:\n{0}".format(set(descr.args.keys()))
            )

    # check that CustomDataLoader indeed interits from the right DataLoader
    if descr.type in DATALOADERS_AS_FUNCTIONS:
        # transform the functions into objects
        assert isinstance(CustomDataLoader, types.FunctionType)
        CustomDataLoader = AVAILABLE_DATALOADERS[descr.type].from_fn(CustomDataLoader)
    else:
        if not issubclass(CustomDataLoader, AVAILABLE_DATALOADERS[descr.type]):
            raise ValueError("DataLoader does't inherit from the specified dataloader: {0}".
                             format(AVAILABLE_DATALOADERS[descr.type].__name__))
    logger.info('successfully loaded the dataloader {} from {}'.
                format(dataloader, os.path.normpath(os.path.join(dataloader_dir, descr.defined_as))))

    # enrich the original dataloader class with description
    Dl = CustomDataLoader._add_description_factory(descr)
    # add other fields
    Dl.source = source
    Dl.source_dir = dataloader_dir
    return Dl


# simplify - alias
get_dataloader_factory = get_dataloader

# NOTE: the dataloaders need to be ordered in a way they inherit from each other
# e.g. child can't appear before the parent in the list below
# TODO this could be automatically checked in the future
AVAILABLE_DATALOADERS = OrderedDict([
    ("PreloadedDataset", PreloadedDataset),
    ("Dataset", Dataset),
    ("BatchDataset", BatchDataset),
    ("SampleIterator", SampleIterator),
    ("SampleGenerator", SampleGenerator),
    ("BatchIterator", BatchIterator),
    ("BatchGenerator", BatchGenerator)])

DATALOADERS_AS_FUNCTIONS = ["PreloadedDataset", "SampleGenerator", "BatchGenerator"]



from kipoi_utils.external.related.mixins import RelatedConfigMixin
from kipoi_utils.external.related.fields import StrSequenceField
import os
import kipoi
import related
import logging
from collections import OrderedDict

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

_MODEL_ENV_DB = None


def get_model_env_db():
    global _MODEL_ENV_DB
    if _MODEL_ENV_DB is None:
        _MODEL_ENV_DB = EnvDb(kipoi.config._env_db_path)
    return _MODEL_ENV_DB


def reload_model_env_db():
    global _MODEL_ENV_DB
    if _MODEL_ENV_DB is not None:
        del _MODEL_ENV_DB
    _MODEL_ENV_DB = None
    get_model_env_db()


@related.mutable
class EnvCreateArgs(RelatedConfigMixin):
    model = StrSequenceField(str, required=True)
    source = related.StringField(required=True)
    dataloader = StrSequenceField(str, default=[], required=False)
    env = related.StringField(default=None, required=False)
    gpu = related.BooleanField(default=False, required=False)
    tmpdir = related.StringField(default=None, required=False)
    vep = related.BooleanField(default=False, required=False)


@related.mutable
class EnvDbEntry(RelatedConfigMixin):
    """
    Compatible Models explicitly and the possible (intermediate) groups (derive from kipoi/models/shared/envs/models.yaml (can be model group too) or models in the group if it was a group: sub_models = list_subcomponents(parsed_model, parsed_source, "model") - TODO - check kipoi.cli.env.merge_deps)
    Store all the parsed args
    Store the conda version
    Store the kipoi version
    Timestamp
    successful
    cli_path
    """
    conda_version = related.StringField(required=True)
    kipoi_version = related.StringField(required=True)
    timestamp = related.FloatField(required=True)
    compatible_models = related.SequenceField(str, required=True)
    create_args = related.ChildField(EnvCreateArgs,
                                     required=True)
    successful = related.BooleanField(default=False, required=False)
    cli_path = related.StringField(default=None, required=False)


class EnvDb:
    def __init__(self, db_path):
        from tinydb import TinyDB
        self.entries = []
        self.db = TinyDB(db_path)
        # There is a main problem with querying the way it is done here. The content of the DB is only loaded once
        # and therefore the DB is not used as a DB and the object must the re-instantiated to update the entries.
        self._load_entries()

    def _load_entries(self):
        self.entries = []
        # load everything into memory
        for db_entry in self.db:
            try:
                self.entries.append(EnvDbEntry.from_config(db_entry))
            except Exception as e:
                logger.warning("Could not load entry with cli path {0} due to: {1}. "
                            "Skipping...".format(str(db_entry), str(e)))

    def get_entry_by_model(self, model_name, only_most_recent=True, only_valid=False):
        # iterate over all the entries and select the ones where the model_name is part of one of the listed models
        # For checking split the model_name by "/" as well as the env-compatible model names and then check equality.
        # Select the one with the most recent timestamp
        norm_name = lambda x: x.lstrip("/").rstrip("/")
        norm_model_name = norm_name(model_name)
        query_model_tk_len = len(norm_model_name.split("/"))
        sel_entries = {}
        for entry in self.get_all(only_valid=only_valid):
            pre_sel = [m for m in entry.compatible_models if model_name in m]
            sel = [m for m in pre_sel if "/".join(norm_name(m).split("/")[:query_model_tk_len]) == norm_model_name]
            if len(sel) != 0:
                sel_entries[entry.timestamp] = entry
        ordered_entries = OrderedDict([(k, sel_entries[k]) for k in sorted(list(sel_entries.keys()))][::-1])

        if only_most_recent:
            if len(ordered_entries) == 0:
                return None
            else:
                return list(ordered_entries.values())[0]
        else:
            return list(ordered_entries.values())

    def get_all_unfinished(self):
        unfinished = []
        for e in self.entries:
            if not e.successful or e.cli_path is None or not os.path.exists(e.cli_path):
                unfinished.append(e)
        return unfinished

    def db_remove_unfinished(self):
        [self.remove(e) for e in self.get_all_unfinished()]

    def get_all(self, only_valid=False):
        entries = self.entries
        if only_valid:
            invalid = self.get_all_unfinished()
            entries = [e for e in entries if e not in invalid]
        return entries

    def remove(self, entry):
        self.entries = [e for e in self.entries if e != entry]

    def append(self, entry):
        self.entries.append(entry)

    def save(self):
        self.db.truncate()
        for entry in self.entries:
            self.db.insert(entry.get_config())

    def __del__(self):
        self.db.close()



"""Plugin function
"""
import importlib
import pkgutil
import pandas as pd

PLUGINS = [
    {"plugin": "kipoi_veff",
     "url": "https://github.com/kipoi/kipoi-veff",
     "cli": True,
     "description": "Variant effect prediction"},
    {"plugin": "kipoi_interpret",
     "url": "https://github.com/kipoi/kipoi-interpret",
     "cli": True,
     "description": "Model interpretation using feature importance scores like ISM, grad*input or DeepLIFT."},
]


def is_installed(package):
    """Check if a python package (Kipoi plugin) is installed

    Args:
      package (str): package/plugin name

    Returns:
      True if package/plugin is installed
    """
    return pkgutil.find_loader(package) is not None


def load_plugin(plugin):
    """Get the plugin module
    """
    return importlib.import_module(plugin)


def list_installed(cli_support=False):
    """
    Args:
       cli_support: if True, only packages with CLI
           support will be listed
    Returns:
      list of installed plugin names
    """
    def cli_filter(plugin):
        if cli_support:
            return plugin['cli']
        else:
            return True

    return [plugin['plugin']
            for plugin in PLUGINS
            if is_installed(plugin['plugin']) and cli_filter(plugin)]


def list_plugins():
    """List available plugins

    Returns:
      pd.DataFrame with columns:
        - plugin (str) - plugin name - should correspond to the python name
        - installed (bool) - True if the plugin is installed
        - cli (bool) - True if the package supports the command-line interface
        - description (str) - short description
        - url (str) - url of the plugin
    """
    df = pd.DataFrame.from_dict(PLUGINS)
    df['installed'] = pd.Series([is_installed(plugin)
                                 for plugin in df.plugin])
    return df[['plugin', 'installed', 'cli', 'description', 'url']]


def is_plugin(name):
    """Test if name is a pluin
    """
    return name in [p['plugin'] for p in PLUGINS]


def get_model_yaml_parser(plugin):
    """
    Returns:
      Related class
    """
    return load_plugin(plugin).ModelParser


def get_dataloader_yaml_parser(plugin):
    """
    Returns:
      Related class
    """
    return load_plugin(plugin).DataloaderParser


def get_cli_fn(plugin):
    """
    Returns:
      CLI function
    """
    return load_plugin(plugin).cli_main


def plugin2cli(plugin):
    return plugin.replace("kipoi_", "")


def get_plugin_cli_fns():
    return {plugin2cli(plugin): get_cli_fn(plugin)
            for plugin in list_installed(cli_support=True)}


def get_plugin_help():
    header = "\n    # - plugin commands:"
    dfp = list_plugins()
    plugin_msg = []
    for i in range(len(dfp)):
        p = dfp.iloc[i]
        if p.installed and p.cli:
            plugin_msg.append("    {:<17}{}".format(plugin2cli(p.plugin), p.description))
    if len(plugin_msg) == 0:
        plugin_msg = ["No plugins installed. "
                      "Install them using pip install <plugin>. Available plugins: {0}".
                      format(", ".join(dfp.plugin))]
    return "\n".join([header] + plugin_msg) + "\n"



from __future__ import absolute_import
from __future__ import print_function

from abc import ABCMeta, abstractmethod, abstractproperty
import sys
import os
import six
import subprocess
import logging
from collections import OrderedDict
from kipoi_utils.utils import unique_list, lfs_installed, get_file_path, cd, list_files_recursively, is_subdir, relative_path
import pandas as pd
import kipoi
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


# --------------------------------------------

# main functions

def list_subcomponents(component, source, which="model"):
    """List all the available submodels

    Args:
      model: model name or a subname: e.g. instaead of
        Model1/CTCF we can give Model1 and then all the sub-models would be included
      source: model source
    """
    src = kipoi.get_source(source)
    if src._is_component(component, which):
        return [component]
    else:
        return [x for x in src._list_components(which)
                if x.startswith(component) and "/template" not in x]


def get_model_descr(model, source="kipoi"):
    """Get model description

    # Arguments
        model: model's relative path/name in the source. 2nd column in the `kipoi.list_models() `pd.DataFrame`.
        source: Model source. 1st column in the `kipoi.list_models()` `pd.DataFrame`.
    """
    return kipoi.config.get_source(source).get_model_descr(model)


def get_dataloader_descr(dataloader, source="kipoi"):
    """Get dataloder description

    # Arguments
        datalaoder: dataloader's relative path/name in the source. 2nd column in the `kipoi.list_dataloader() `pd.DataFrame`.
        source: Model source. 1st column in the `kipoi.list_models()` `pd.DataFrame`.
    """
    return kipoi.config.get_source(source).get_dataloader_descr(dataloader)


# --------------------------------------------

# helper

def get_component_file(component_dir, which="model", raise_err=True):
    # TODO - if component_dir has an extension, then just return that file path
    return get_file_path(component_dir, which, extensions=[".yml", ".yaml"], raise_err=raise_err)


def list_yamls_recursively(root_dir, basename):
    return [os.path.dirname(x) for x in list_files_recursively(root_dir, basename, suffix='y?ml')]


def list_softlink_realpaths(root_dir):
    for root, dirnames, filenames in os.walk(root_dir):
        for name in filenames:
            fname = os.path.join(root, name)
            if os.path.islink(fname):
                yield os.path.realpath(fname)


def load_component_descr(component_dir, which="model"):
    """Return the parsed yaml file
    """
    from kipoi.specs import ModelDescription, DataLoaderDescription

    fname = get_component_file(os.path.abspath(component_dir), which, raise_err=True)

    with cd(os.path.dirname(fname)):
        if which == "model":
            return ModelDescription.load(fname)
        elif which == "dataloader":
            return DataLoaderDescription.load(fname)
        else:
            raise ValueError("which needs to be from {'model', 'dataloader'}")


def list_softlink_dependencies(component_dir, source_path):
    """List dependencies of a directory

    Returns a set
    """
    return {relative_path(f, source_path) if os.path.isdir(f)
            else relative_path(os.path.dirname(f), source_path)
            for f in list_softlink_realpaths(component_dir)
            if is_subdir(f, source_path)}


def to_namelist(nested_dict):
    """no-recursion
    """
    if isinstance(nested_dict, list):
        return [x.name for x in nested_dict]
    elif isinstance(nested_dict, dict):
        return list(nested_dict)
    else:
        return nested_dict.name


def list_models_by_group(df, group_filter=""):
    """Get a list of models by a group

    Args:
      df: Dataframe returned by list_models()
      group_filter, str: A relative path to the model group used to subset
        model list.

    Returns:
      a pd.DataFrame with columns (or None in case no groups are found):
        - group - name of the sub-group
        - N_models
        - is_group
        - authors
        - contributors
        - doc - ?
        - type (list of types)
        - tags - sum
    """
    # df = self.list_models()
    # add slashes
    if group_filter == "":
        group = "/"
    else:
        group = "/" + group_filter + "/"
    df = df[df.model.str.contains("^" + group[1:])].copy()
    # df['parent_group'] = group[1:]
    df['model'] = df.model.str.replace("^" + group[1:], "")
    df['is_group'] = df.model.str.contains("/")
    if not df.is_group.any():
        return None

    df = df.join(df.model.str.split("/", n=1, expand=True).rename(columns={0: "group", 1: "child"}))

    def n_subgroups(ch):
        if ch.str.contains("/").sum() > 0:
            return len(ch[ch.str.contains("/")].str.split("/", n=1, expand=True).iloc[:, 0].unique())
        else:
            return 0

    def fn(x):
        # remove the templates
        return pd.Series(OrderedDict([
            ("N_models", x.shape[0]),
            ("N_subgroups", n_subgroups(x.child.fillna(""))),
            ("is_group", x.is_group.any()),
            ("authors", unique_list([author for authors in x.authors
                                     for author in authors])),
            ("contributors", unique_list([contributor for contributors in x.contributors
                                          for contributor in contributors])),
            ("veff_score_variants", x.veff_score_variants.any()),
            ("type", unique_list([t for t in x.type])),
            ("license", unique_list([l for l in x.license])),
            ("cite_as", unique_list([c for c in x.cite_as if c is not None])),
            ("tags", unique_list([tag for tags in x.tags
                                  for tag in tags])),
        ]))

    return df.groupby("group").apply(fn).reset_index()


class LocalComponentGroup(object):
    """Abstraction for a folder containing the following files:

    - model-template.yaml
    - models.tsv
    - (optional) dataloader.yaml
    - ...
    """

    def __init__(self, component_template_yaml, models_tsv, which='model'):
        from jinja2 import Template
        # read the yaml file as a string
        self.which = which
        assert self.which in ['model', 'dataloader']

        self.component_template_yaml = component_template_yaml
        with open(self.component_template_yaml, "r") as f:
            template_str = f.read()
            if sys.version_info[0] == 2:
                template_str = template_str.decode("utf-8")
            self.template = Template(template_str)

        self.models_tsv = models_tsv
        self.df = pd.read_csv(models_tsv, sep='\t', comment='#')

        if 'model' not in self.df:
            raise ValueError("Column 'model' has to exist in {}. Make "
                             "also sure the tsv file is correctly formatted".format(models_tsv))

        # assert each model occurs once
        assert len(self.df.model.duplicated()) == len(self.df.model)

    def get_model_row(self, model):
        if model not in list(self.df.model):
            raise ValueError("model {} not found in {}".format(model, list(self.df.model)))
        return self.df[self.df.model == model].iloc[0].to_dict()

    def _is_component(self, component):
        return component in self._list_components()

    def _list_components(self):
        return list(self.df.model)

    def _get_component_descr(self, component):
        from kipoi.specs import ModelDescription, DataLoaderDescription

        # render the template
        rendered_yaml = self.template.render(**self.get_model_row(component))

        if self.which == 'model':
            return ModelDescription.from_string(rendered_yaml)
        elif self.which == 'dataloader':
            return DataLoaderDescription.from_string(rendered_yaml)
        else:
            raise ValueError("Unknown component {}".format(self.which))

    # --------------------------------------------
    # class methods

    @classmethod
    def is_group(cls, path, which='model'):
        models_tsv = os.path.join(path, 'models.tsv')
        yaml_template = get_component_file(path, which + '-template', raise_err=False)
        if yaml_template is None:
            return False
        return os.path.exists(models_tsv) and os.path.exists(yaml_template)

    @classmethod
    def group_path(cls, component_path, which='model'):
        """Get the path of the group given the component

        if None is returned, then the component was not found
        """
        tmp_path = component_path
        while tmp_path not in ['', '/']:
            if cls.is_group(tmp_path, which):
                return tmp_path
            else:
                tmp_path = os.path.dirname(tmp_path)
        return None

    @classmethod
    def load(cls, path, which='model'):
        """Load the component group from the directory
        """
        models_tsv = os.path.join(path, 'models.tsv')
        yaml_template = get_component_file(path, which + '-template')

        if not os.path.exists(models_tsv):
            raise ValueError("models.tsv doesn't exist in model group {}".format(path))
        if not os.path.exists(yaml_template):
            raise ValueError("{}-template.yaml doesn't exist in model group {}".format(which, path))
        return cls(yaml_template, models_tsv, which)

# --------------------------------------------

# Source abstract class


class Source(object):

    __metaclass__ = ABCMeta

    # --------------------------------------------
    # implemented by childs
    @abstractproperty
    def TYPE(self):
        pass

    @abstractmethod
    def _list_components(self, which="model"):
        """List available models a strings
        """
        pass

    @abstractmethod
    def _pull_component(self, component, which="model"):
        """Pull/update the model locally
        """
        return

    @abstractmethod
    def _is_component(self, component, which='model'):
        """Returns True if the component exists
        """
        return

    @abstractmethod
    def _get_component_dir(self, component, which='model'):
        """Get component directory
        """
        return

    @abstractmethod
    def _get_component_download_dir(self, component, which='model'):
        """Get component dedicated download directory
        """
        return

    @abstractmethod
    def _get_component_descr(self, component, which="model"):
        """Given the component name, return the description
        """
        pass

    @abstractmethod
    def get_config(self):
        pass

    @abstractmethod
    def get_group_name(self, component, which='model'):
        pass
    # --------------------------------------------

    def assert_is_component(self, component, which='model'):
        if not self._is_component(component, which):
            raise ValueError("{} {} doesn't exist".format(which, component))

    def pull_model(self, model):
        self._pull_component(model, "model")
        return self.get_model_dir(model)

    def pull_dataloader(self, dataloader):
        self._pull_component(dataloader, "dataloader")
        return self.get_dataloader_dir(dataloader)

    def get_model_dir(self, model):
        return self._get_component_dir(model, 'model')

    def get_dataloader_dir(self, model):
        return self._get_component_dir(model, 'dataloader')

    def get_model_download_dir(self, model):
        return self._get_component_download_dir(model, 'model')

    def get_dataloader_download_dir(self, dataloader):
        return self._get_component_download_dir(dataloader, 'dataloader')

    def list_models(self):
        """List all the models as a data.frame
        """
        def dict2df_dict(d, model):
            return OrderedDict([
                ("model", model),
                # ("name", d.info.name),
                ("version", d.info.version),
                ("authors", d.info.authors),
                ("contributors", d.info.contributors),
                ("doc", d.info.doc),
                ("type", d.type),
                ("inputs", to_namelist(d.schema.inputs)),
                ("targets", to_namelist(d.schema.targets)),
                ("veff_score_variants", "variant_effects" in d.postprocessing),
                ("license", d.info.license),
                ("cite_as", d.info.cite_as),
                ("trained_on", d.info.trained_on),
                ("training_procedure", d.info.training_procedure),
                ("tags", d.info.tags),
            ])

        df = pd.DataFrame([dict2df_dict(self.get_model_descr(model), model)
                           for model in self._list_components("model")])
        if len(df):
            # filter all template models
            return df[~df.model.str.contains("/template$")]
        else:
            return df

    def list_dataloaders(self):
        """List all the models as a data.frame
        """
        def dict2df_dict(d, dataloader):
            return OrderedDict([
                ("dataloader", dataloader),
                # ("name", d.info.name),
                ("version", d.info.version),
                ("authors", d.info.authors),
                ("doc", d.info.doc),
                ("type", d.type),
                ("inputs", to_namelist(d.get_output_schema().inputs)),
                ("targets", to_namelist(d.get_output_schema().targets)),
                ("license", d.info.license),
                ("tags", d.info.tags),
            ])

        df = pd.DataFrame([dict2df_dict(self.get_dataloader_descr(dataloader), dataloader)
                           for dataloader in self._list_components("dataloader")])
        # filter all template models
        return df[~df.dataloader.str.contains("/template$")]

    def list_models_by_group(self, group_filter=""):
        """Get a list of models by a group

        Args:
          group_filter, str: A relative path to the model group used to subset
            model list.

        Returns:
          a pd.DataFrame with columns (or None in case no groups are found):
            - group - name of the sub-group
            - N_models
            - is_group
            - authors
            - contributors
            - doc - ?
            - type (list of types)
            - tags - sum
        """
        return list_models_by_group(self.list_models(), group_filter)

    def get_model_descr(self, model):
        return self._get_component_descr(model, which="model")

    def get_dataloader_descr(self, dataloader):
        return self._get_component_descr(dataloader, which="dataloader")

    @classmethod
    def from_config(cls, config):
        assert config.pop("type") == cls.TYPE

        return cls(**config)

    def __repr__(self):
        conf = self.get_config()
        cls_name = self.__class__.__name__
        conf.pop("type")

        kwargs = ', '.join('{0}={1}'.format(k, repr(v))
                           for k, v in six.iteritems(conf))
        return "{0}({1})".format(cls_name, kwargs)


# --------------------------------------------

# individual source implementations


class LocalSource(Source):

    TYPE = "local"

    def __init__(self, local_path=None, name=None):
        """Local files
        """
        self.name = name
        if local_path is not None:
            self._local_path = os.path.join(os.path.realpath(local_path), '')  # add trailing slash

            # load the config file
            config_file = os.path.join(self._local_path, 'config.yaml')
            if os.path.exists(config_file):
                from kipoi.specs import SourceConfig
                self.config = SourceConfig.load(config_file)

                if not self.config.dependencies.all_installed(verbose=False):
                    import colorlog
                    print(colorlog.escape_codes['red'])
                    print("WARNING: Dependencies for model source '{}' stored at local_path {} not satisfied.: \n---".
                          format(self.name, self._local_path))
                    self.config.dependencies.all_installed(verbose=True)
                    print("---\ninstall or update the missing packages")
                    print(colorlog.escape_codes['reset'])
                    print("Note: If you don't want to auto_update the model source, \n"
                          "add `auto_update: False` to ~/.kipoi/config.yaml\n")
            else:
                self.config = None
        else:
            # undetermined local path
            self._local_path = None
            self.config = None
        self.component_yaml_list = None
        self.component_group_list = None

    @property
    def local_path(self):
        if self._local_path is None:
            return os.getcwd()
        else:
            return self._local_path

    def _list_component_yamls(self, which='model'):
        return list_yamls_recursively(self.local_path, which)  # , skip='downloaded|.*_files')

    def _list_component_groups(self, which='model'):
        return {tdir: LocalComponentGroup.load(os.path.join(self.local_path, tdir), which)
                for tdir in list_yamls_recursively(self.local_path, which + '-template')}  # , skip='downloaded|.*_files')}

    def cache_component_list(self, force=False):
        if force or self.component_yaml_list is None or self.component_group_list is None:
            self.component_yaml_list = dict(model=self._list_component_yamls(which="model"),
                                            dataloader=self._list_component_yamls(which="dataloader"))
            self.component_group_list = dict(model=self._list_component_groups(which="model"),
                                             dataloader=self._list_component_groups(which="dataloader"))

    def _list_components(self, which="model"):
        self.cache_component_list(force=self._local_path is None)
        return self.component_yaml_list[which] + [os.path.join(k, c)
                                                  for k, grp in six.iteritems(self.component_group_list[which])
                                                  for c in grp._list_components()]

    def get_group_name(self, component, which='model'):
        component = os.path.normpath(component)

        if self.component_group_list is not None:
            # already cached
            for k in self.component_group_list[which]:
                if component.startswith(os.path.join(k, "")):
                    return k
            return None
        else:
            group_path = LocalComponentGroup.group_path(os.path.join(self.local_path, component), which)
            if group_path is None:
                return None
            else:
                return relative_path(group_path, self.local_path)

    def _is_nongroup_component(self, component, which):
        path = os.path.join(self.local_path, os.path.normpath(component))
        if get_component_file(path, which=which, raise_err=False) is not None:
            return True
        else:
            return False

    def _get_component_dir(self, component, which='model'):
        component = os.path.normpath(component)

        self.assert_is_component(component, which)
        # special case: component can be outside of the root directory
        if self._is_nongroup_component(component, which):
            return os.path.join(self.local_path, os.path.normpath(component))
        else:
            k = self.get_group_name(component, which)
            assert k is not None
            return os.path.join(self.local_path, k)

    def _get_component_download_dir(self, component, which='model', name=None):
        component = os.path.normpath(component)

        if name is None:
            name = which
        insert_path = os.path.join("downloaded", '{}_files'.format(name))

        # special case: component can be outside of the root directory
        if self._is_nongroup_component(component, which):
            return os.path.join(self.local_path, os.path.normpath(component), insert_path)
        else:
            k = self.get_group_name(component, which)
            if k is None and which == 'dataloader':
                # fallback: get model's download directory
                return self._get_component_download_dir(component, which='model', name='dataloader')
            if k is None:
                raise ValueError("Couldn't get {} download_dir. Model"
                                 " or group doesn't exist for {}".format(which, component))
            return os.path.join(self.local_path, k, insert_path, relative_path(component, k))

    def _is_component(self, component, which="model"):
        component = os.path.normpath(component)
        if self._is_nongroup_component(component, which):
            # it contains a {which}.y?ml
            return True
        else:
            # it's present in one of the groups

            k = self.get_group_name(component, which)
            if k is not None:
                # check that it's indeed found in one of the components
                if self.component_group_list is not None:
                    # already cached
                    return component in self._list_components(which)
                else:
                    grp = LocalComponentGroup.load(os.path.join(self.local_path, k), which)
                    return grp._is_component(relative_path(component, k))
            else:
                return False

    def _get_component_descr(self, component, which="model"):
        component = os.path.normpath(component)
        self.assert_is_component(component, which)

        if self._is_nongroup_component(component, which):
            # component has an explicit yaml file

            # TODO - move into the component directory when loading
            return load_component_descr(os.path.join(self.local_path, component), which)
        else:
            k = self.get_group_name(component, which)
            if k is None:
                raise ValueError("{} {} doesn't exist".format(which, component))
            else:
                if self.component_group_list is not None:
                    # already cached
                    return self.component_group_list[which][k]._get_component_descr(relative_path(component, k))
                else:
                    grp = LocalComponentGroup.load(os.path.join(self.local_path, k), which)
                    return grp._get_component_descr(relative_path(component, k))

    def _pull_component(self, component, which="model"):
        self.assert_is_component(component, which)

    def get_config(self):
        return OrderedDict([("type", self.TYPE),
                            ("local_path", self.local_path)])


class GitSource(Source):

    TYPE = "git"

    def __init__(self, remote_url, local_path, auto_update=True, use_lfs=False, name=None):
        """Git Source
        """
        self.name = name
        self.remote_url = remote_url
        self.local_path = os.path.join(os.path.realpath(local_path), '')  # add trailing slash
        self.local_source = LocalSource(self.local_path, name=name)

        self.auto_update = auto_update
        self._pulled = False
        self.use_lfs = use_lfs
        if self.use_lfs:
            lfs_installed(raise_exception=False)

    def clone(self, depth=1):
        """Clone the self.remote_url into self.local_path

        Args:
          depth: --depth argument to git clone. If None, clone the whole history.
        """
        if os.path.exists(self.local_path) and os.listdir(self.local_path):
            raise IOError("Directory {0} already exists and is non-empty".
                          format(self.local_path))

        logger.info("Cloning {remote} into {local}".
                    format(remote=self.remote_url,
                           local=self.local_path))
        cmd = ["git", "clone"]
        if depth is not None:
            cmd.append("--depth={0}".format(depth))
        cmd.append(self.remote_url)
        cmd.append(self.local_path)
        subprocess.call(cmd,
                        env=dict(os.environ, GIT_LFS_SKIP_SMUDGE="1"))
        self._pulled = True

        if os.path.exists(os.path.join(self.local_path, ".gitattributes")):
            if not self.use_lfs:
                logger.info(".gitattributes detected in {}. Using git-lfs".format(self.local_path))
            self.use_lfs = True
            lfs_installed(raise_exception=True)

    def pull_source(self):
        """Pull/update the source
        """
        if not os.path.exists(self.local_path) or not os.listdir(self.local_path):
            return self.clone()

        if not self.auto_update:
            logger.warning("Pulling source even though auto_update=False")

        logger.info("Update {0}".
                    format(self.local_path))
        subprocess.call(["git",
                         "pull"],
                        cwd=self.local_path,
                        env=dict(os.environ, GIT_LFS_SKIP_SMUDGE="1"))

        if os.path.exists(os.path.join(self.local_path, ".gitattributes")):
            if not self.use_lfs:
                logger.info(".gitattributes detected in {}. Using git-lfs".format(self.local_path))
            self.use_lfs = True
            lfs_installed(raise_exception=True)

        self._pulled = True

    def _commit_checkout(self, commit):
        """Checkout a particular commit
        """
        logger.info("Update {0}".
                    format(self.local_path))
        subprocess.call(["git",
                         "reset",
                         "--hard",
                         commit],
                        cwd=self.local_path,
                        env=dict(os.environ, GIT_LFS_SKIP_SMUDGE="1"))

    def _list_components(self, which="model"):
        if not self._pulled and self.auto_update:
            self.pull_source()
        return self.local_source._list_components(which)

    def _get_component_dir(self, component, which='model'):
        if not self._pulled and self.auto_update:
            self.pull_source()
        return self.local_source._get_component_dir(component, which)

    def _get_component_download_dir(self, component, which='model'):
        if not self._pulled and self.auto_update:
            self.pull_source()
        return self.local_source._get_component_download_dir(component, which)

    def _pull_component(self, component, which="model"):
        if not self._pulled and self.auto_update:
            self.pull_source()

        component_dir = self.local_source._get_component_dir(component, which)

        if self.use_lfs:
            # the only call to git-lfs -> pulling specific sub-files

            lfs_installed(raise_exception=True)
            # get a list of directories to source (relative to the local_path)
            softlink_dirs = list(list_softlink_dependencies(component_dir, self.local_path))
            # pull these softlinks
            for pull_dir in [component, relative_path(component_dir, self.local_path)] + softlink_dirs:
                cmd = ["git-lfs",
                       "pull",
                       "-I {0}/**".format(pull_dir)]
                logger.info(" ".join(cmd))
                subprocess.call(cmd,
                                cwd=self.local_path,
                                env=dict(os.environ, GIT_LFS_SKIP_SMUDGE="1"))

        return self.local_source._pull_component(component, which)

    def _is_component(self, component, which="model"):
        if not self._pulled and self.auto_update:
            self.pull_source()
        return self.local_source._is_component(component, which)

    def get_group_name(self, component, which='model'):
        return self.local_source.get_group_name(component, which)

    def _get_component_descr(self, component, which="model"):
        if not self._pulled and self.auto_update:
            self.pull_source()
        return self.local_source._get_component_descr(component, which)

    def get_config(self):
        return OrderedDict([("type", self.TYPE),
                            ("remote_url", self.remote_url),
                            ("local_path", self.local_path),
                            ("auto_update", self.auto_update),
                            ("use_lfs", self.use_lfs)])


class GitLFSSource(GitSource):
    TYPE = 'git-lfs'

    def __init__(self, remote_url, local_path, auto_update=True, name=None):
        """Git-LFS Source
        """
        super(GitLFSSource, self).__init__(remote_url=remote_url,
                                           local_path=local_path,
                                           auto_update=auto_update,
                                           use_lfs=True,
                                           name=name)


class GithubPermalinkSource(Source):

    TYPE = "github-permalink"

    def __init__(self, local_path, name=None):
        """Local files
        """
        self.name = name
        self.local_path = os.path.join(os.path.realpath(local_path), '')  # add trailing slash

    @classmethod
    def _parse_url(cls, url):
        """Map github url to local directory
        """
        github_url = "https://github.com/"
        if not url.startswith(github_url):
            raise ValueError("url of the permalink: {0} doesn't start with {1}".format(url, github_url))
        url_dir = url[len(github_url):]
        if "/tree/" not in url_dir:
            raise ValueError("'/tree/' missing in the url {0}. Typical github format " +
                             "is github.com/<user>/<repo>/tree/<commit>/<directory>")
        url_dir = url_dir.replace("/tree", "")

        user, repo, commit, model = url_dir.split("/", 3)
        model = model.rstrip("/")
        return user, repo, commit, model

    def _list_components(self, which="model"):
        # Same as for local source
        return []  # list_yamls_recursively(self.local_path, which)

    def get_lfs_source(self, component):
        user, repo, commit, model = self._parse_url(component)
        remote_url = "https://github.com/{user}/{repo}.git".format(user=user, repo=repo)
        lfs_source = GitSource(remote_url, os.path.join(self.local_path, user, repo, commit),
                               auto_update=False,  # Don't git-pull
                               use_lfs=True)
        self._pulled = True  # actually not required due to auto_update=False

        if not os.path.exists(lfs_source.local_path) or not os.listdir(lfs_source.local_path):
            # clone the repository
            lfs_source.clone(depth=None)
            lfs_source._commit_checkout(commit)
        return lfs_source

    def _get_component_dir(self, component, which='model'):
        user, repo, commit, model = self._parse_url(component)
        return self.get_lfs_source(component)._get_component_dir(model, which='model')

    def _get_component_download_dir(self, component, which='model'):
        user, repo, commit, model = self._parse_url(component)
        return self.get_lfs_source(component)._get_component_download_dir(model, which='model')

    def _pull_component(self, component, which="model"):
        user, repo, commit, model = self._parse_url(component)
        self.get_lfs_source(component)._pull_component(model, which=which)

    def _is_component(self, component, which="model"):
        user, repo, commit, model = self._parse_url(component)
        return self.get_lfs_source(component)._is_component(model, which)

    def get_group_name(self, component, which='model'):
        user, repo, commit, model = self._parse_url(component)
        return self.get_lfs_source(component).get_group_name(component, which)

    def _get_component_descr(self, component, which="model"):
        user, repo, commit, model = self._parse_url(component)
        return self.get_lfs_source(component)._get_component_descr(model, which)

    def get_config(self):
        return OrderedDict([("type", self.TYPE),
                            ("local_path", self.local_path)])


# --------------------------------------------
# all available models
source_classes = [GitLFSSource, GitSource, LocalSource, GithubPermalinkSource]


def load_source(config, name):
    """Load the source from config
    """
    type_cls = {cls.TYPE: cls for cls in source_classes}
    if config["type"] not in type_cls:
        raise ValueError("config['type'] needs to be one of: {0}".
                         format(list(type_cls.keys())))

    if config['type'] == 'git-lfs':
        # local path is already checked out and .gitattributes doesn't exist
        # -> the repo is not an actuall git-lfs
        if os.path.exists(os.path.join(config['local_path'], '.git')) \
           and not os.path.exists(os.path.join(config['local_path'], '.gitattributes')):
            # if the .gitattributes doesn't exist for the local path

            if not config['remote_url'].endswith("kipoi/models.git"):
                # only display this message for remote url's other than kipoi/models
                logger.info(".gitattributes not found for git-lfs source: {} stored at {}".
                            format(config['remote_url'], config['local_path']))
                logger.info("Using 'type: git'")
            config['type'] = 'git'

    # add name
    config['name'] = name
    cls = type_cls[config["type"]]
    return cls.from_config(config)



"""test-source command
"""
from __future__ import absolute_import
from __future__ import print_function
from io import open

import argparse
import logging
import os
import re
import sys
from copy import deepcopy

from colorlog import escape_codes, default_log_colors

import kipoi
from kipoi.cli.env import conda_env_name, SPECIAL_ENV_PREFIX
from kipoi_conda import get_kipoi_bin, env_exists, remove_env, _call_command
from kipoi.sources import list_softlink_dependencies
from kipoi_utils.utils import list_files_recursively, read_txt, get_file_path, cd

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def modified_files(git_range, source_folder, relative=True):
    """
    Returns files under the models dir that have been modified within the git
    range. Filenames are returned with the `source_folder` included.

    Args:
      git_range : list or tuple of length 1 or 2
          For example, ['00232ffe', '10fab113'], or commonly ['master', 'HEAD']
          or ['master']. If length 2, then the commits are provided to `git diff`
          using the triple-dot syntax, `commit1...commit2`. If length 1, the
          comparison is any changes in the working tree relative to the commit.
      source_folder : str
          Root of the model source/git repo
      relative=True: return the relative path
    """
    assert isinstance(git_range, list)
    cmds = ['diff', '--name-only'] + git_range

    with cd(source_folder):
        code, lines = _call_command("git", cmds, use_stdout=True,
                                    return_logs_with_stdout=True)

    assert code == 0
    modified = [os.path.join(source_folder, line)
                for line in lines]

    # exclude files that were deleted in the git-range
    existing = list(filter(os.path.exists, modified))

    # if the only diff is that files were deleted, we can have ['model/'], so
    # filter on existing *files*
    existing = list(filter(os.path.isfile, existing))
    if relative:
        return [os.path.relpath(f, source_folder)
                for f in existing]
    else:
        return existing


def all_models_to_test(src):
    """Returns a list of models to test

    By default, this method returns all the model. In case a model group has a
    `test_subset.txt` file present in the group directory, then testing is only
    performed for models listed in `test_subset.txt`.

    Args:
      src: Model source
    """
    txt_files = list_files_recursively(src.local_path, "test_subset", "txt")

    exclude = []
    include = []
    for x in txt_files:
        d = os.path.dirname(x)
        exclude += [d]
        include += [os.path.join(d, l)
                    for l in read_txt(os.path.join(src.local_path, x))]

    # try to load every model extra included -- will get tested downstream
    # for m in include:
    #     src.get_model_descr(m)

    models = src.list_models().model
    for excl in exclude:
        models = models[~models.str.startswith(excl)]
    return list(models) + include


def create_model_env(model_name, source_name, env_name, vep=False, verbose=False):
    """kipoi test ...

    Args:
      model_name (str)
      source_name: source name
    """
    if env_exists(env_name):
        logger.info("Environment {0} exists. Removing it.".format(env_name))
        remove_env(env_name)

    # TODO - if the model is a Keras model, print the Keras config file
    # and note which config file got used

    # create the model test environment
    cmd = "kipoi"
    args = ["env", "create",
            "--source", source_name,
            "--env", env_name,
            model_name]
    if verbose:
        args.append("--verbose")
    if vep:
        # Add --vep to environment installation
        args.insert(-1, "--vep")
    returncode = _call_command(cmd, args, use_stdout=True)
    assert returncode == 0


def test_model(model_name, source_name, env_name, batch_size,
               vep=False, create_env=True, verbose=False):
    """kipoi test ...

    Args:
      model_name (str)
      source_name: source name
    """
    if create_env:
        create_model_env(model_name, source_name, env_name, vep, verbose=verbose)

    # run the tests in the environment
    cmd = get_kipoi_bin(env_name)
    args = ["test",
            "--batch_size", str(batch_size),
            "--source", source_name,
            model_name]
    # New, modified path for conda. Source activate namely does the following:
    # - CONDA_DEFAULT_ENV=${env_name}
    # - CONDA_PREFIX=${env_path}
    # - PATH=$conda_bin:$PATH
    new_env = os.environ.copy()
    new_env['PATH'] = os.path.dirname(cmd) + os.pathsep + new_env['PATH']
    returncode, logs = _call_command(cmd, args, use_stdout=True,
                                     return_logs_with_stdout=True,
                                     env=new_env
                                     )
    assert returncode == 0

    # detect WARNING in the output log
    warn = 0
    for line in logs:
        warn_start = escape_codes[default_log_colors['WARNING']] + \
            'WARNING' + escape_codes['reset']
        if line.startswith(warn_start):
            logger.error("Warning present: {0}".format(line))
            warn += 1
    if warn > 0:
        raise ValueError("{0} warnings were observed for model {1}".
                         format(warn, model_name))


def restrict_models_to_test(all_models, source, git_range):
    """Subset all_models to the ones with changed files

    Args:
      all_models: list of all models to test
      source: used source
      git_range: a tuple of (from, to) master branches

    Inspired by bioconda-utils ... --git-range
    1. check
    """
    modified = modified_files(git_range, source.local_path, relative=True)

    def dependency_modified(model_name, modified_files):
        """Test if the dependency was modified
        """
        def contains_any(x, file_list):
            """Test if the directory x contains any files from file_list
            """
            for y in file_list:
                if y.startswith(x + os.sep) or x == y:
                    return True
            return False

        # if model is from a model group, test if any of the files were modified
        group_name = source.get_group_name(model_name, which='model')
        if group_name is not None:
            # Test if any files required by the group were modified
            if contains_any(group_name, modified_files):
                return True

        # get all the dependency directories
        dep_dirs = list_softlink_dependencies(os.path.join(source.local_path,
                                                           model_name),
                                              source.local_path)

        for model_dir in [model_name] + list(dep_dirs):
            if contains_any(model_dir, modified_files):
                return True
        return False

    return [x for x in all_models if dependency_modified(x, modified)]


def rm_env(env_name):
    """Alias for remove_env
    """
    from kipoi.env_db import get_model_env_db
    if env_exists(env_name):
        logger.info("Removing environment: {0}".
                    format(env_name))
        remove_env(env_name)
        # remove from db
        db = get_model_env_db()
        db_entries = [e for e in db.get_all() if e.create_args.env == env_name]
        [db.remove(e) for e in db_entries]
        db.save()


def get_batch_size(cfg, model_name, default=4):
    """Get the default batch size to test
    """
    if cfg is not None and model_name in cfg.test.constraints:
        bs = cfg.test.constraints[model_name].batch_size
        if bs is not None:
            if bs != default:
                logger.info("Using a different batch_size ({0})".
                            format(bs) +
                            " for model {0} than the default ({1})".
                            format(model_name, default))
            return bs
    return default


def get_common_env(model_name, model_envs):
    """Get the common environment name
    """
    env_name = None
    for env in model_envs:
        m_group = deepcopy(model_name)
        while m_group:
            if m_group in model_envs[env]:
                env_name = env
                break
            # try to get the environment matching any specific parent
            m_group = os.path.dirname(m_group)
        if env_name is not None:
            break
    return env_name


def cli_test_source(command, raw_args):
    """Runs test on the model
    """
    assert command == "test-source"
    # setup the arg-parsing
    parser = argparse.ArgumentParser('kipoi {}'.format(command),
                                     description='Test models in model source')
    parser.add_argument('source', default="kipoi",
                        help='Which source to test')
    parser.add_argument('--git-range', nargs='+',
                        help='''Git range (e.g. commits or something like
                        "master HEAD" to check commits in HEAD vs master, or just "HEAD" to
                        include uncommitted changes). All models modified within this range will
                        be tested.''')
    parser.add_argument('-n', '--dry_run', action='store_true',
                        help='Dont run model testing')
    parser.add_argument('-b', '--batch_size', default=4, type=int,
                        help='Batch size')
    parser.add_argument('-x', '--exitfirst', action='store_true',
                        help='exit instantly on first error or failed test.')
    parser.add_argument('-k', default=None,
                        help='only run tests which match the given substring expression')
    parser.add_argument('-c', '--clean_env', action='store_true',
                        help='clean the environment after running.')
    parser.add_argument('--vep', action='store_true',
                        help='Install the vep dependency.')
    parser.add_argument('--common_env', action='store_true',
                        help='Test models in common environments.')
    parser.add_argument('--all', action='store_true',
                        help="Test all models in the source")
    parser.add_argument('-v', '--verbose', action='store_true',
                        help="Increase output verbosity. Show conda stdout during env installation.")
    args = parser.parse_args(raw_args)
    # --------------------------------------------
    source = kipoi.get_source(args.source)
    all_models = all_models_to_test(source)
    if args.k is not None:
        all_models = [x for x in all_models if re.match(args.k, x)]

    if len(all_models) == 0:
        logger.info("No models found in the source")
        sys.exit(1)
    if args.all:
        test_models = all_models
        logger.info('Testing all models:\n- {0}'.
                    format('\n- '.join(test_models)))
    else:
        test_models = restrict_models_to_test(all_models,
                                              source,
                                              args.git_range)
        if len(test_models) == 0:
            logger.info("No model modified according to git, exiting.")
            sys.exit(0)
        logger.info('{0}/{1} models modified according to git:\n- {2}'.
                    format(len(test_models), len(all_models),
                           '\n- '.join(test_models)))
    # Sort the models alphabetically
    test_models = sorted(test_models)

    # Parse the repo config
    cfg_path = get_file_path(source.local_path, "config",
                             extensions=[".yml", ".yaml"],
                             raise_err=False)
    if cfg_path is not None:
        cfg = kipoi.specs.SourceConfig.load(cfg_path, append_path=False)
        logger.info("Found config {0}:\n{1}".format(cfg_path, cfg))
    else:
        cfg = None

    if args.dry_run:
        logger.info("-n/--dry_run enabled. Skipping model testing and exiting.")
        sys.exit(0)

    # TODO - make sure the modes are always tested in the same order?
    #        - make sure the keras config doesn't get cluttered

    # Test common environments
    if args.common_env:
        logger.info("Installing common environmnets")
        import yaml
        models_yaml_path = os.path.join(source.local_path, SPECIAL_ENV_PREFIX, "models.yaml")
        if not os.path.exists(models_yaml_path):
            logger.error("{} doesn't exists when installing the common environment".format(models_yaml_path))
            sys.exit(1)
        model_envs = yaml.safe_load(open(os.path.join(source.local_path, SPECIAL_ENV_PREFIX, "models.yaml"), "r", encoding="utf-8"))

        test_envs = {get_common_env(m, model_envs) for m in test_models if get_common_env(m, model_envs) is not None}

        if len(test_envs) == 0:
            logger.info("No common environments to test")
            sys.exit(0)

        logger.info("Instaling environments covering the following models: \n{}".format(yaml.dump(model_envs)))
        for env in test_envs:
            if env_exists(env):
                logger.info("Common environment already exists: {}. Skipping the installation".format(env))
            else:
                logger.info("Installing environment: {}".format(env))
                create_model_env(os.path.join(SPECIAL_ENV_PREFIX, env), args.source, env, vep=args.vep)

    logger.info("Running {0} tests..".format(len(test_models)))
    failed_models = []
    for i in range(len(test_models)):
        m = test_models[i]
        print('-' * 20)
        print("{0}/{1} - model: {2}".format(i + 1,
                                            len(test_models),
                                            m))
        print('-' * 20)
        try:
            if not args.common_env:
                # Prepend "test-" to the standard kipoi env name
                env_name = conda_env_name(m, source=args.source)
                env_name = "test-" + env_name
                # Test
                test_model(m, args.source, env_name,
                           get_batch_size(cfg, m, args.batch_size), args.vep,
                           create_env=True, verbose=args.verbose)
            else:
                # figure out the common environment name
                env_name = get_common_env(m, model_envs)
                if env_name is None:
                    # skip is none was found
                    logger.info("Common environment not found for {}".format(m))
                    continue
                # ---------------------------
                # Test
                print("test_model...")
                test_model(m, args.source, env_name,
                           get_batch_size(cfg, m, args.batch_size), args.vep,
                           create_env=False, verbose=args.verbose)
        except Exception as e:
            logger.error("Model {0} failed: {1}".format(m, e))
            failed_models += [m]
            if args.exitfirst:
                if args.clean_env and not args.common_env:
                    rm_env(env_name)
                sys.exit(1)
        finally:
            if args.clean_env and not args.common_env:
                rm_env(env_name)
    print('-' * 40)
    if failed_models:
        logger.error("{0}/{1} tests failed for models:\n- {2}".
                     format(len(failed_models),
                            len(test_models),
                            "\n- ".join(failed_models)))
        sys.exit(1)

    logger.info('All tests ({0}) passed'.format(len(test_models)))
